import numpy as np
from numpy.linalg import inv
from itertools import combinations
from scipy.stats import norm as norm_dist
from scipy.stats import gamma as gamma_dist
from scipy.stats import bernoulli as bernoulli_dist
from scipy.special import polygamma, digamma, gamma, beta
from scipy.optimize import minimize

class ibp_model:
    '''
    Factor analysis model with positive sparse continuous factors.  This version includes include an intercept term.
    The precision of observed variables (xi) is fixed.
    
    ***** FINISH UPDATING *****
    
    Parameters
    ----------
    x: array
        Matrix of observed data.  Rows are variables and columns are observations.
    alpha: float, optional
        Parameter of the Indian buffet process.
    omega: float, optional
        Prior precision of A.
    prior_mS: float, optional
        Prior mean of S.
    prior_nS: float, optional
        Prior "strength" (virtual number of observations) for S.
    xi: float, optional
        Precision of observed variables.
    p: integer, optional
        Number of latent factors used in the finite apporximation to the Indian buffet process.
    include_intercept: logical, optional
        If True (default) then an intercept term is included.  If False then it is not.
    
    Notes
    -----
    The continuous latent factor component (S) has a gamma distribution, and thus is strictly positive. 
    *****
    '''
    
    def __init__(self, x, alpha = 2.0, omega = 0.2, prior_mS = 1.0, prior_nS = 5.0, xi = 1, p = 4, include_intercept = True):
        """
        Parameters
        ----------
        x: array
            Matrix of observed data.  Rows are variables and columns are observations.
        alpha: float, optional
            Parameter of the Indian buffet process.
        omega: float, optional
            Prior precision of A.
        xi: float, optional
            Precision of observed variables.
        p: integer, optional
            Number of latent factors used in the finite apporximation to the Indian buffet process.
        include_intercept: logical, optional
            If True (default) then an intercept term is included.  If False then it is not.
        """
        # ----- SET UP SOME VARIABLES -----
        self.n = x.shape[1] # number of observations
        self.m = x.shape[0] # number of observed variables
        self.p = p
        self.mu = 1 - np.isnan(x) # = 1 if not missing, = 0 if missing CHECK THIS
        self.x = x
        self.x[np.isnan(x)] = 0 # this is an arbitrary value to make computations work
        self.tilde_x = self.mu*x
        self.tilde_xsq = self.mu*x**2
        self.tilde_x_sum = np.sum(self.tilde_x)
        self.tilde_xsq_sum = np.sum(self.tilde_xsq)
        self.tilde_n = self.mu.sum(axis = 1)
        self.alpha = alpha
        self.omega = omega
        self.prior_aS = prior_nS/2 # one conventional prior hyperparameter for S
        self.prior_bS = prior_nS/(2*prior_mS) # the other conventional prior hyperparameter for S
        # lists for storing the ELBO and its components at each iteration
        self.E_log_lik_list = []
        self.E_log_prior_list = []
        self.entropy_list = []
        self.elbo_list = []
        self.rng = np.random.default_rng() # for generating random permutations
        self.xi = xi # fixed precision of observed variables
        self.include_intercept = include_intercept
        # indices used for computing E_q[(A_{j,:} z_{:,i})^2]
        index_combos = np.array(list(combinations(range(self.p + self.include_intercept), 2)))
        self.kneql = index_combos[:,0]
        self.lneqk = index_combos[:,1]
        self.n_combos = self.lneqk.shape[0]

        # ----- INITIALIZE VARIATIONAL HYPERPARAMETERS -----
        
        # initialize A's hyperparameters
        self.mA = self.rng.normal(loc = 0.0, scale = 1/np.sqrt(self.omega), size = [self.m, self.p + self.include_intercept])
        self.vA = 0.1*np.ones([self.m, self.p + self.include_intercept])
        self.E_Asq = self.mA**2 + self.vA
        # initialize u's hyperparameters
        self.nu = self.rng.uniform(low = 0.2, high = 0.8, size = [self.p, self.n])
        # initialize S's hyperparameters
        self.aS = self.prior_aS*np.ones([self.p, self.n])
        self.bS = self.prior_bS*np.ones([self.p, self.n])
        self.E_S = self.aS/self.bS
        # initialize pi's hyperparameters
        nu_sum = self.nu.sum(axis = 1)
        self.tau1 = self.alpha/self.p + nu_sum
        self.tau2 = 1 + self.n - nu_sum
        self.E_pi = self.tau1/(self.tau1 + self.tau2)
    
    def fit(self, stop_threshold = 1e-05, max_iter = 50, min_iter = 30):
        """
        Parameters
        ----------
        stop_threshold: float, optional
            Relative change in the ELBO at which the optimization should stop.
        max_iter: integer, optional
            Maximum number of iterations to run the optimization.
        min_iter: integer, optional
            Minimum number of iterations to run the optimization.
        """
        
        relative_elbo_change = 1.0 # this is an arbitrary value for the first iteration
        itr = 0 # counter for iterations
        continue_loop = True
        int_elbo = -np.Inf
        
        while continue_loop:
            # ----- UPDATE VARIATIONAL HYPERPARAMETERS -----
            update_s = True
            if update_s:
                # update S's hyperparameters (in random order)
                for i in self.rng.permutation(self.n):
                    for k in self.rng.permutation(self.p):
                        # constants used for computing the ELBO gradient
                        lval = list(range(0, k)) + list(range(k+1, self.mA.shape[1])) # l : l =/= k
                        C1 = self.nu[k,i]*np.sum(self.xi*self.mu[:,i]*self.mA[:,k]*(self.x[:,i] - self.mA[:,lval]@self.nu[lval,i]))
                        C2 = -0.5*self.nu[k,i]*np.sum(self.xi*self.mu[:,i]*self.E_Asq[:,k])
                        p0 = np.log(self.aS[k,i])
                        p1 = np.log(self.bS[k,i])
                        for s in range(5):
                            new_a = np.exp(p0)
                            new_b = np.exp(p1)
                            # compute the ELBO gradient (log scale)
                            gradient0 = new_a*((1/new_b)*(C1 - self.prior_bS) + ((2*new_a + 1)/(new_b**2))*C2 + (self.prior_aS - new_a)*polygamma(1, new_a) + 1) # partial derivative of the ELBO wrt p0 = log(a)
                            gradient1 = new_b*((new_a/(new_b**2))*(self.prior_bS - C1) - ((2*new_a*(new_a + 1))/(new_b**3))*C2 - self.prior_aS/new_b) # partial derivative of the ELBO wrt p1 = log(b)
                            # do the gradient ascent step (log scale)
                            p0 += 0.005*gradient0
                            p1 += 0.005*gradient1
                        # update aS and bS
                        self.aS[k,i] = np.exp(p0)
                        self.bS[k,i] = np.exp(p1)
                # compute mean and variance of S
                self.mS = self.aS/self.bS
                self.vS = self.aS/self.bS**2
                        
                        # function to compute -ELBO as a function of aS[k,i] and bS[k,i], and its gradient
                       # def neg_ELBO_and_gradient(pars):
                       #     new_a = np.exp(pars[0])
                       #     new_b = np.exp(pars[1])
                       #     new_full_aS = self.aS.copy()
                       #     new_full_aS[k,i] = new_a
                       #     new_full_bS = self.bS.copy()
                       #     new_full_bS[k,i] = new_b
                            
                            # compute the ELBO
                       #     E_log_lik = self.compute_E_log_lik(aS = new_full_aS, bS = new_full_bS)
                       #     E_log_prior = self.compute_E_log_prior(aS = new_full_aS, bS = new_full_bS)
                       #     entropy = self.compute_entropy(aS = new_full_aS, bS = new_full_bS)
                       #     ELBO = E_log_lik + E_log_prior + entropy
                            
                            # compute the gradient of the ELBO
                       #     gradient = np.zeros(2)
                       #     gradient[0] = new_a*((1/new_b)*(C1 - self.prior_bS) + ((2*new_a + 1)/(new_b**2))*C2 + (self.prior_aS - new_a)*polygamma(1, new_a) + 1) # partial derivative of the ELBO wrt pars[0]
                       #     gradient[1] = new_b*((new_a/(new_b**2))*(self.prior_bS - C1) - ((2*new_a*(new_a + 1))/(new_b**3))*C2 - self.prior_aS/new_b) # partial derivative of the ELBO wrt pars[1]
                            
                       #     return -ELBO, -gradient
                            
                        # use the BFGS algorithm to find to optimal values of aS[k,i] and bS[k,i] (on the log scale)
                        #result = minimize(fun = neg_ELBO_and_gradient,
                        #                  #x0 = np.array([np.log(self.aS[k,i]), np.log(self.bS[k,i])]),
                        #                  x0 = np.ones(2),
                        #                  method = 'BFGS',
                        #                  jac = True) # specifies that the gradient is contained in the function
                        
                        #self.aS[k,i] = np.exp(result['x'][0])
                        #self.bS[k,i] = np.exp(result['x'][1])
            
            update_u = True
            if update_u:
                # update u's hyperparameters (in random order)
                for i in self.rng.permutation(self.n):
                    for k in self.rng.permutation(self.p):
                        E_z = self.compute_E_z()
                        lval = list(range(0, k)) + list(range(k+1, self.mA.shape[1])) # latent factor indices excluding k, i.e. l : l =/= k
                        theta_ki = np.sum(self.mu[:,i]*self.xi*( self.mA[:,k]*self.mS[k,i]*(self.x[:,i] - self.mA[:,lval]@E_z[lval,i]) - 0.5*(self.mA[:,k]**2 + self.vA[:,k])*(self.mS[k,i]**2 + self.vS[k,i]) )) + digamma(self.tau1[k]) - digamma(self.tau2[k])
                        self.nu[k,i] = 1/(1 + np.exp(-theta_ki))
            
            # update pi's hyperparameters
            nu_sum = self.nu.sum(axis = 1)
            self.tau1 = self.alpha/self.p + nu_sum
            self.tau2 = 1 + self.n - nu_sum
            self.E_pi = self.tau1/(self.tau1 + self.tau2)
            
            # update A's hyperparameters
            E_z = self.compute_E_z()
            E_z_sq = self.compute_E_z_sq()
            for k in self.rng.permutation(self.mA.shape[1]):
                lval = list(range(0, k)) + list(range(k+1, self.mA.shape[1])) # l : l =/= k
                for j in range(self.m):
                    self.vA[j,k] = (self.xi*np.sum(self.mu[j,:]*E_z_sq[k,:]) + self.omega)**(-1)
                    self.mA[j,k] = self.vA[j,k]*self.xi*np.sum(self.mu[j,:]*E_z[k,:]*(self.x[j,:] - self.mA[j,lval]@E_z[lval,:]))
            self.E_Asq = self.mA**2 + self.vA

            # ----- COMPUTE ELBO -----
            self.E_log_lik_list += [self.compute_E_log_lik()]
            self.E_log_prior_list += [self.compute_E_log_prior()]
            self.entropy_list += [self.compute_entropy()]
            self.elbo_list += [self.E_log_lik_list[-1] + self.E_log_prior_list[-1] + self.entropy_list[-1]]
            # ----- FINISH CURRENT ITERATION -----      
            itr += 1 # increment the iteration counter
            if itr < min_iter: # automatically continue if below the minimum number of iterations
                continue_loop = True
            else: # otherwise, continue if ELBO change is below threshold and we haven't yet hit the iteration limit
                relative_elbo_change = np.abs((self.elbo_list[-2] - self.elbo_list[-1])/self.elbo_list[-2])
                continue_loop = (relative_elbo_change > stop_threshold) and (itr < max_iter)
                
    def compute_E_z(self, nu = None, aS = None, bS = None):
        if aS is None:
            aS = self.aS
        if bS is None:
            bS = self.bS
        if nu is None:
            nu = self.nu
        E_z = nu*aS/bS
        if self.include_intercept:
            E_z = np.append(E_z, np.ones([1, self.n]), axis = 0)
        return E_z
            
    def compute_E_z_sq(self, nu = None, aS = None, bS = None):
        if aS is None:
            aS = self.aS
        if bS is None:
            bS = self.bS
        if nu is None:
            nu = self.nu
        E_S_sq = (aS/(bS**2))*(aS + 1)
        E_z_sq = nu*E_S_sq
        if self.include_intercept:
            E_z_sq = np.append(E_z_sq, np.ones([1, self.n]), axis = 0)
        return E_z_sq
                
    def compute_E_Az_sq(self, nu = None, aS = None, bS = None, mA = None, vA = None):
        '''
        Compute E_q[(A_{j,:} z_{:,i})^2].
        '''
        if aS is None:
            aS = self.aS
        if bS is None:
            bS = self.bS
        if nu is None:
            nu = self.nu
        if mA is None:
            mA = self.mA
        if vA is None:
            vA = self.vA
        E_z = self.compute_E_z(nu, aS, bS)
        E_z_sq = self.compute_E_z_sq(nu, aS, bS)
        E_Az_sq = np.zeros([self.m, self.n])
        for i in range(self.n):
            for j in range(self.m):
                E_Az_sq[j,i] = (vA[j,:] + mA[j,:]**2)@E_z_sq[:,i] + 2*np.sum(mA[j,self.kneql]*mA[j,self.lneqk]*E_z[self.kneql,i]*E_z[self.lneqk,i])
        return E_Az_sq
    
    def compute_E_log_lik(self, aS = None, bS = None, nu = None, tau1 = None, tau2 = None, mA = None, vA = None):
        if aS is None:
            aS = self.aS
        if bS is None:
            bS = self.bS
        if nu is None:
            nu = self.nu
        if tau1 is None:
            tau1 = self.tau1
        if tau2 is None:
            tau2 = self.tau2
        if mA is None:
            mA = self.mA
        if vA is None:
            vA = self.vA
        E_z = self.compute_E_z()
        E_Az_sq = self.compute_E_Az_sq()
        term1 = 0
        for j in range(self.m): # 1st term of E_log_lik
            term1 += self.xi*np.sum(self.mu[j,:]*( (mA[j,:]@E_z)*self.x[j,:] -0.5*self.x[j,:]**2 - 0.5*E_Az_sq[j,:] ))
        term2 = 0.5*np.sum( self.tilde_n*( np.log(self.xi) - np.log(2*np.pi) ) ) # 2nd term of E_log_lik
        return term1 + term2 # this is split into two terms for convenience
    
    def compute_E_log_prior(self, aS = None, bS = None, nu = None, tau1 = None, tau2 = None, mA = None, vA = None):
        if aS is None:
            aS = self.aS
        if bS is None:
            bS = self.bS
        if nu is None:
            nu = self.nu
        if tau1 is None:
            tau1 = self.tau1
        if tau2 is None:
            tau2 = self.tau2
        if mA is None:
            mA = self.mA
        if vA is None:
            vA = self.vA
        E_log_A_prior = -0.5*self.omega*np.sum(vA + mA**2) + 0.5*self.m*self.p*(np.log(self.omega) - np.log(2*np.pi))
        E_log_u_prior = np.sum( (digamma(tau1) - digamma(tau2))*np.sum(nu, axis = 1) + self.n*(digamma(tau2) - digamma(tau1 + tau2)) )
        E_log_S_prior = self.n*self.p*( self.prior_aS*np.log(self.prior_bS) - np.log(gamma(self.prior_aS)) ) + np.sum( (self.prior_aS - 1)*(digamma(aS) - np.log(bS)) - self.prior_bS*aS/bS )
        E_log_pi_prior = (self.alpha/self.p - 1)*np.sum(digamma(tau1) - digamma(tau1 + tau2)) + self.p*np.log(self.alpha/self.p)
        return E_log_A_prior + E_log_u_prior + E_log_S_prior + E_log_pi_prior
    
    def compute_entropy(self, aS = None, bS = None, nu = None, tau1 = None, tau2 = None, mA = None, vA = None):
        if aS is None:
            aS = self.aS
        if bS is None:
            bS = self.bS
        if nu is None:
            nu = self.nu
        if tau1 is None:
            tau1 = self.tau1
        if tau2 is None:
            tau2 = self.tau2
        if mA is None:
            mA = self.mA
        if vA is None:
            vA = self.vA
        H_A = norm_dist.entropy(mA, np.sqrt(vA)).sum()
        H_u = bernoulli_dist.entropy(nu + 1e-16).sum()
        H_S = gamma_dist.entropy(aS + 1e-16, bS + 1e-16).sum()
        H_pi = gamma_dist.entropy(tau1 + 1e-16, tau2 + 1e-16).sum()
        return H_A + H_u + H_S + H_pi


import numpy as np
from itertools import combinations
from scipy.stats import norm as norm_dist
from scipy.stats import bernoulli as bernoulli_dist
from scipy.special import digamma, gamma, beta

class ibp_model:
    '''
    Factor analysis model with discrete factors.  This version includes include an intercept term.
    The precision of observed variables (xi) is fixed.
    
    Parameters
    ----------
    x: array
        Matrix of observed data.  Rows are variables and columns are observations.
    alpha: float, optional
        Parameter of the Indian buffet process.
    omega: float, optional
        Prior precision of A.
    xi: float, optional
        Precision of observed variables.
    p: integer, optional
        Number of latent factors used in the finite approximation to the Indian buffet process.
    include_intercept: logical, optional
        If True (default) then an intercept term is included.  If False then it is not.
    
    Notes
    -----
    *****
    '''
    
    def __init__(self, x, alpha = 2.0, omega = 0.2, xi = 1, p = 4, include_intercept = True):
        """
        Parameters
        ----------
        x: array
            Matrix of observed data.  Rows are variables and columns are observations.
        alpha: float, optional
            Parameter of the Indian buffet process.
        omega: float, optional
            Prior precision of A.
        xi: float, optional
            Precision of observed variables.
        p: integer, optional
            Number of latent factors used in the finite apporximation to the Indian buffet process.
        include_intercept: logical, optional
            If True (default) then an intercept term is included.  If False then it is not.
        """
        # ----- SET UP SOME VARIABLES -----
        self.n = x.shape[1] # number of observations
        self.m = x.shape[0] # number of observed variables
        self.p = p
        self.mu = 1 - np.isnan(x) # = 1 if not missing, = 0 if missing CHECK THIS
        self.x = x
        self.x[np.isnan(x)] = 0 # this is an arbitrary value to make computations work
        self.tilde_x = self.mu*x
        self.tilde_xsq = self.mu*x**2
        self.tilde_x_sum = np.sum(self.tilde_x)
        self.tilde_xsq_sum = np.sum(self.tilde_xsq)
        self.tilde_n = self.mu.sum(axis = 1)
        self.alpha = alpha
        self.omega = omega
        self.elbo_list = [] # list for storing ELBO values at each iteration
        self.rng = np.random.default_rng() # for generating random permutations
        self.xi = xi # fixed precision of observed variables
        self.include_intercept = include_intercept
        # indices used for computing E_q[(A_{j,:} z_{:,i})^2]
        index_combos = np.array(list(combinations(range(self.p + self.include_intercept), 2)))
        self.kneql = index_combos[:,0]
        self.lneqk = index_combos[:,1]
        # a variable used in several calculations
        self.E_Az_sq = np.zeros([self.m, self.n])

        # ----- INITIALIZE VARIATIONAL HYPERPARAMETERS -----
        
        # initialize A's hyperparameters
        #self.mA = self.rng.uniform(low = -0.05, high = 0.05, size = [self.m, self.p + self.include_intercept])
        self.mA = self.rng.normal(loc = 0.0, scale = 1/np.sqrt(self.omega), size = [self.m, self.p + self.include_intercept])
        self.vA = 2.0*np.ones([self.m, self.p + self.include_intercept])        
        # initialize z's hyperparameters
        #self.nu = self.rng.uniform(low = 0.1, high = 0.25, size = [self.p, self.n])
        #self.nu = self.rng.uniform(low = 0.2, high = 0.8, size = [self.p, self.n])
        self.nu = self.rng.uniform(low = 0.01, high = 0.99, size = [self.p, self.n])
        self.update_nu_plus() # if include_intercept then this adds a row of 1s for the intercept, otherwise it is equal to nu
        # initialize pi's hyperparameters
        nu_sum = self.nu.sum(axis = 1)
        self.tau1 = self.alpha/self.p + nu_sum
        self.tau2 = 1 + self.n - nu_sum
        self.E_pi = self.tau1/(self.tau1 + self.tau2)
    
    def fit(self, stop_threshold = 1e-05, max_iter = 50, min_iter = 30):
        """
        Parameters
        ----------
        stop_threshold: float, optional
            Relative change in the ELBO at which the optimization should stop.
        max_iter: integer, optional
            Maximum number of iterations to run the optimization.
        min_iter: integer, optional
            Minimum number of iterations to run the optimization.
        """
        
        relative_elbo_change = 1.0 # this is an arbitrary value for the first iteration
        itr = 0 # counter for iterations
        continue_loop = True
        int_elbo = -np.Inf        
        
        while continue_loop:
            # ----- UPDATE VARIATIONAL HYPERPARAMETERS -----
            
            # update A's hyperparameters
            self.update_nu_plus()
            for j in range(self.m):
                for k in range(self.mA.shape[1]):
                    lval = list(range(0, k)) + list(range(k+1, self.mA.shape[1])) # latent factor indices excluding k, i.e. l : l =/= k
                    self.vA[j,k] = (self.xi*np.sum(self.mu[j,:]*self.nu_plus[k,:]) + self.omega)**(-1)
                    self.mA[j,k] = self.vA[j,k]*self.xi*np.sum(self.nu_plus[k,:]*(self.tilde_x[j,:] - self.mu[j,:]*(self.mA[j,lval]@self.nu_plus[lval,:])))

            # update z's hyperparameters (in random order)
            for i in self.rng.permutation(self.n):
                for k in range(self.p):
                    self.update_nu_plus()
                    lval = list(range(0, k)) + list(range(k+1, self.mA.shape[1])) # latent factor indices excluding k, i.e. l : l =/= k
                    #theta_ki = np.sum(self.mu[:,i]*self.xi*( self.mA[:,k]*(self.x[:,i] - self.mA[:,lval]@self.nu_plus[lval,i]) - 0.5*(self.vA[:,k] + self.mA[:,k]**2) )) + digamma(self.tau1[k]) - digamma(self.tau2[k])
                    theta_ki = self.xi*( np.sum(self.mu[:,i]*self.mA[:,k]*self.x[:,i]) - np.sum(self.mu[:,i]*self.mA[:,k]*(self.mA[:,lval]@self.nu_plus[lval,i])) - 0.5*np.sum(self.mu[:,i]*(self.vA[:,k] + self.mA[:,k]**2)) ) + digamma(self.tau1[k]) - digamma(self.tau2[k])
                    self.nu[k,i] = 1/(1 + np.exp(-theta_ki))

            # update pi's hyperparameters
            nu_sum = self.nu.sum(axis = 1)
            self.tau1 = self.alpha/self.p + nu_sum
            self.tau2 = 1 + self.n - nu_sum
            self.E_pi = self.tau1/(self.tau1 + self.tau2)

            # ----- COMPUTE ELBO -----
            elbo = self.compute_elbo()
            # ----- FINISH CURRENT ITERATION -----            
            self.elbo_list += [elbo] # record the current ELBO
            itr += 1 # increment the iteration counter
            if itr < min_iter: # automatically continue if below the minimum number of iterations
                continue_loop = True
            else: # otherwise, continue if ELBO change is below threshold and we haven't yet hit the iteration limit
                relative_elbo_change = np.abs((self.elbo_list[-2] - self.elbo_list[-1])/self.elbo_list[-2])
                continue_loop = (relative_elbo_change > stop_threshold) and (itr < max_iter)

    def update_nu_plus(self):
        if self.include_intercept:
            self.nu_plus = np.append(self.nu, np.ones([1, self.n]), axis = 0)
        else:
            self.nu_plus = self.nu
                
    def compute_E_Az_sq(self):
        '''
        Compute E_q[(A_{j,:} z_{:,i})^2].
        '''
        self.update_nu_plus()
        for i in range(self.n):
            for j in range(self.m):
                self.E_Az_sq[j,i] = (self.vA[j,:] + self.mA[j,:]**2)@self.nu_plus[:,i] + 2*np.sum(self.mA[j,self.kneql]*self.mA[j,self.lneqk]*self.nu_plus[self.kneql,i]*self.nu_plus[self.lneqk,i])
    
    def compute_elbo(self):
        # compute expected log-likelihood
        self.compute_E_Az_sq()
        term1 = 0
        for j in range(self.m): # 1st term of E_log_lik
            term1 += self.xi*( self.mA[j,:]@(self.nu_plus@self.tilde_x[j,:].T) - 0.5*self.tilde_xsq_sum - 0.5*np.sum(self.mu[j,:]*self.E_Az_sq[j,:]) )
        term2 = 0.5*np.sum( self.tilde_n*( np.log(self.xi) - np.log(2*np.pi) ) ) # 2nd term of E_log_lik
        E_log_lik = term1 + term2 # this is split into two terms for convenience

        # compute expected log-prior
        E_log_A_prior = -0.5*self.omega*np.sum(self.vA + self.mA**2) - 0.5*self.m*self.p*(np.log(2*np.pi) - np.log(self.omega))
        E_log_z_prior = np.sum( (digamma(self.tau1) - digamma(self.tau2))*np.sum(self.nu, axis = 1) + self.n*(digamma(self.tau2) - digamma(self.tau1 + self.tau2)) )
        E_log_pi_prior = (self.alpha/self.p - 1)*np.sum(digamma(self.tau1) - digamma(self.tau1 + self.tau2)) + self.p*np.log(self.alpha/self.p)
        E_log_prior = E_log_A_prior + E_log_z_prior + E_log_pi_prior

        # compute entropy
        H_A = norm_dist.entropy(self.mA, np.sqrt(self.vA)).sum()
        H_z = bernoulli_dist.entropy(self.nu).sum()
        H_pi = np.sum(np.log(np.maximum(beta(self.tau1, self.tau2), 1e-320)) - (self.tau1 - 1)*digamma(self.tau1) - (self.tau2 - 1)*digamma(self.tau2) + (self.tau1 + self.tau2 - 2)*digamma(self.tau1 + self.tau2)) # the scipy.stats entropy method doesn't appear to work well with certain parameter values
        entropy = H_A + H_z + H_pi

        return E_log_lik + E_log_prior + entropy


import numpy as np

def latent_profile(n_profiles = 4, m = 10, n = 500, separation = 2.0, mu = None, pi = None, sample_z = True, seed = None):
    '''
    Simulate data from a latent profile analysis (LPA) model.
    
    Parameters
    ----------
    n_profiles: int, optional
        Number of latent profiles. Ignored if either mu (latent profile means)
        or pi (latent profile base rates) is manually specified.
    m: int, optional
        Number of observed (indicator) variables. Ignored if either mu 
        (latent profile means) is manually specified.
    n: int, optional
        Number of participants (observations) to simulate.
    mu: array or None, optional
        Optionally specify profile means as a m x n_profiles array.
        If None (default) then profile means are sampled from a
        a normal distribution.
    separation: float, optional
        Variance for generating mu (latent profile means).
    pi: array or None, optional
        Base rates (prior probabilities) for latent profiles.
        If None (the default) then these are sampled from a flat
        Dirichlet distribution.
    sample_z: logical, optional
        If True (default) then values of z (profile membership) are sampled
        from the profile base rates (pi). If False, then profiles will be
        deterministically assigned so that their proportions will be as close
        as possible to the profile base rates, with any extra observations being
        assigned to the last profile.
    seed: int or None, optional
        If an integer, then this is the seed for random number generation.
        Using the same seed for repeated calls to the function will always
        produce the same simulated data. If None (the default) then repeated
        calls to the function will produce different simulated data.
        
    Output
    ------
    pi: Profile base rates (prior probabilties).
    mu: Profile means.
    z: Latent profile membership for each participant (observation).
    z_1hot: 1-hot encoding of z (= 1 for the correct profile, = 0 for all others).
    x: Observed (indicator) variables.
    n_profiles: Number of latent profiles.
    m: Number of observed (indicator) variables.
    n: Number of participants (observations).
    
    Notes
    -----
    The residual variance of simulated data (x) is set at 1.0.
    
    Because of this fact and the fact that profile means (mu) are
    drawn from a normal distribution with mean 0, the following relationship
    holds between Mahalanobis distance and the 'separation' parameter (variance
    for drawing profile means):
    
    separation = (average squared Mahalanobis distance)/(2*m)
    
    Similarly,
    
    separation = (average squared Cohen's d between profiles)/2
    '''
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed = seed)

    # overwrite n_profiles if pi is manually specified
    if not pi is None:
        n_profiles = pi.shape[0]
    # overwrite n_profiles and m if mu is manually specified
    if not mu is None:
        m = mu.shape[0]
        n_profiles = mu.shape[1]
    
    # sample profile base rates if not specified (sort highest to lowest)
    if pi is None:
        pi = np.sort(rng.dirichlet(alpha = np.ones(n_profiles)))[::-1]

    # sample profile centers if not specified
    if mu is None:
        mu = rng.normal(loc = 0.0, scale = np.sqrt(separation), size = [m, n_profiles])

    # assign observations to profiles
    if sample_z == True:
        z = np.zeros(n, dtype = int)
        for i in range(n):
            z[i] = rng.choice(np.arange(n_profiles), p = pi)
    else:
        number_in_each = np.floor(pi*n).astype(int)
        z = np.array(sum([number_in_each[t]*[t] for t in range(n_profiles)], (n - number_in_each.sum())*[n_profiles-1]))
    z_1hot = np.zeros([n_profiles, n])
    for i in range(n):
        z_1hot[z[i], i] = 1.0

    # sample observed data
    x = np.zeros([m, n])
    for i in range(n):
        x[:,i] = rng.normal(loc = mu[:, z[i]], scale = 1.0)

    return {'pi': pi, 'mu': mu, 'z': z, 'z_1hot': z_1hot, 'x': x, 'n_profiles': n_profiles, 'm': m, 'n': n}
        
def latent_factor(n_factors = 2, m = 10, n = 500, loading_var = 3.0, loadings = None, seed = None):
    '''
    Simulate data from a factor analysis model.
    
    Parameters
    ----------
    n_factors: int, optional
        Number of latent factors.
    m: int, optional
        Number of observed (indicator) variables.
    n: int, optional
        Number of participants (observations) to simulate.
    loading_var: float, optional
        Variance for generating factor loadings. This is not
        used if factor loadings are manually specified.
    loadings: float or None, optional
        Factor loadings. If None (the default), then
        these are randomly generated (see notes for details).
    seed: int or None, optional
        If an integer, then this is the seed for random number generation.
        Using the same seed for repeated calls to the function will always
        produce the same simulated data. If None (the default) then repeated
        calls to the function will produce different simulated data.
        
    Output
    ------
    loadings: Factor loadings.
    f: Individual factor scores.
    x: Observed (indicator) variables.
    n_factors: Number of latent factors.
    m: Number of observed (indicator) variables.
    n: Number of participants (observations).
    
    Notes
    -----
    The residual variance of simulated data (x) is set at 1.0.
    Factor scores (f) are drawn from independent standard normal distributions.
    
    If not specified by the user, loadings are randomly generated from the following process:
        1) Factor loadings are initialized to 0.0.
        2) Observed variables are divided up as evenly as possible among the latent factors,
        so that each observed variable depends on one latent factor.
        3) The factor loadings from step 2 are drawn from a normal distribution with mean 0 and variance loading_var.
    '''
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed = seed)
    
    # sample loadings (if needed)
    if loadings is None:
        size = int(np.round(m/n_factors))
        loadings = np.zeros([m, n_factors])
        for k in range(n_factors):
            start = k*size
            if k < n_factors - 1:
                end = (k + 1)*size
            else:
                end = m
            loadings[range(start, end), k] = rng.normal(loc = 0.0, scale = np.sqrt(loading_var), size = end - start)
                
    # sample latent factor scores
    f = rng.normal(loc = 0.0, scale = 1.0, size = [n_factors, n])
    
    # sample observed data
    x = rng.normal(loc = loadings@f, scale = 1.0, size = [m, n])
    
    return {'loadings': loadings, 'f': f, 'x': x, 'n_factors': n_factors, 'm': m, 'n': n}

def latent_profile_factor(n_profiles = 3, n_factors = 4, m = 10, n = 500, separation = 2.0, pi = None, loading_var = 3.0, loadings = None, seed = 1234):
    '''
    Simulate data from a two-stage process:
    1) Factor scores are generated from latent profiles.
    2) Observed data are generated from factor scores.
    
    Thus, latent profiles specify points in factor space rather than
    observed variable space. This is implicitly the model used when 
    additive scores (as opposed to individual item responses) are fed into
    a latent profile analysis model.
    
    Parameters
    ----------
    n_profiles: int, optional
        Number of latent profiles.
    n_factors: int, optional
        Number of latent factors.
    m: int, optional
        Number of observed (indicator) variables.
    n: int, optional
        Number of participants (observations) to simulate.
    separation: float, optional
        Variance for generating mu (latent profile means).
    pi: float or None, optional
        Base rates (prior probabilities) for latent profiles.
        If None (the default) then these are sampled from a flat
        Dirichlet distribution.
    loading_var: float, optional
        Variance for generating factor loadings. This is not
        used if factor loadings are manually specified.
    loadings: float or None, optional
        Factor loadings. If None (the default), then
        these are randomly generated.
    seed: int or None, optional
        If an integer, then this is the seed for random number generation.
        Using the same seed for repeated calls to the function will always
        produce the same simulated data. If None (the default) then repeated
        calls to the function will produce different simulated data.
        
    Output
    ------
    pi: Profile base rates (prior probabilties).
    mu: Profile means.
    loadings: Factor loadings.
    z: Latent profile membership for each participant (observation).
    z_1hot: 1-hot encoding of z (= 1 for the correct profile, = 0 for all others).
    f: Individual factor scores.
    x: Observed (indicator) variables.
    n_profiles: Number of latent profiles.
    n_factors: Number of latent factors.
    m: Number of observed (indicator) variables.
    n: Number of participants (observations).
    '''
    if seed is None:
        rng = np.random.default_rng()
    else:
        rng = np.random.default_rng(seed = seed)

    # sample profile base rates if not specified (sort highest to lowest)
    if pi is None:
        pi = np.sort(rng.dirichlet(alpha = np.ones(n_profiles)))[::-1]

    # sample profile means (in latent factor space)
    mu = rng.normal(loc = 0.0, scale = np.sqrt(separation), size = [n_factors, n_profiles])

    # assign observations to profiles
    z = np.zeros(n, dtype = int)
    z_1hot = np.zeros([n_profiles, n])
    for i in range(n):
        z[i] = rng.choice(np.arange(n_profiles), p = pi)
        z_1hot[z[i], i] = 1.0

    # sample factor scores based on latent profiles
    f = np.zeros([n_factors, n])
    for i in range(n):
        f[:,i] = rng.normal(loc = mu[:, z[i]], scale = 1.0)

    # sample loadings (if needed)
    if loadings is None:
        size = int(np.round(m/n_factors))
        loadings = np.zeros([m, n_factors])
        for k in range(n_factors):
            start = k*size
            if k < n_factors - 1:
                end = (k + 1)*size
            else:
                end = m
            loadings[range(start, end), k] = rng.normal(loc = 0.0, scale = np.sqrt(loading_var), size = end - start)

    # sample observed data
    x = rng.normal(loc = loadings@f, scale = 1.0, size = [m, n])
    
    return {'pi': pi, 'mu': mu, 'loadings': loadings, 'z': z, 'z_1hot': z_1hot, 'f': f, 'x': x, 'n_profiles': n_profiles, 'n_factors': n_factors, 'm': m, 'n': n}


import numpy as np
import pandas as pd
from scipy.special import betaln, gammaln, multigammaln
from scipy.linalg import inv, det

# These classes are somewhat similar in spirit to this: https://pythonhosted.org/infpy/infpy.exp.html
# I used 1-indexing in the doc, so I'm using pseudo 1-indexing here to avoid mistakes in transcribing the math into Python.

class distribution:
    '''
    Class for exponential family probability distributions.
    
    Notes
    -----
    The LAST axis in each data array etc. indicate observation. This is the opposite
    of many applications.
    
    Here is a summary of the notation conventions used.
    
    natural parameters (of the likelihood): :math:`\theta = \theta_1, \ldots, \theta_p`
    data point: :math:`y_i`
    number of natural parameters (of the likelihood): :math:`p`
    number of data points: :math:`n`
    prior natural hyperparameters: :math:`\tau = \tau_1, \ldots, \tau_p, \tau_{p+1}`
    sufficient statistics: :math:`T(y_i) = T_1(y_i), \ldots, T_p(y_i)`
    posterior natural hyperparameters: :math:`\tau' = \tau_1 + \sum_{i=1}^n T_1(y_i), \ldots, \tau_p + \sum_{i=1}^n T_p(y_i), \tau_{p+1} + n`
    
    Likelihood
    .. math::
    p(y_i | \theta) = \exp\Big( \sum_{j=1}^p \ip{\theta_j, T_j(y_i)} - f(y_i) - g(\theta) \Big) \quad \quad i = 1, 2, \ldots, n
    
    Prior
    .. math::
    p(\theta) = \exp\Big( \sum_{j=1}^p \ip{\tau_j, \theta_j} - \tau_{p+1} g(\theta) - h(\tau) \Big)
    
    Posterior
    .. math::
    p(\theta | y_1, \ldots, y_n) = \exp\Big( \sum_{j=1}^p \ip{\tau_j + \sum_{i=1}^n T_j(y_i), \theta_j} - (\tau_{p+1} + n) g(\theta) - h\big(\tau'(y_1, \ldots, y_n)\big) \Big)
    '''
        
    def tau_prime(self, y):
        '''
        Hyperparameters (tau) plus sufficient statistics of data (T(y)).
        
        Parameters
        ----------
        y: array
            Data.
        '''
        T = self.T(y)
        n = y.shape[-1]
        return [self.tau[j] + np.sum(T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n]
    
    def weighted_tau_prime(self, y, weights):
        '''
        Hyperparameters (tau) plus sufficient statistics of data (T(y)), where each observation
        has a weight.
        
        Parameters
        ----------
        y: array
            Data.
            
        weights: array
            Vector of weights for each data point. Each should be in the range [0, 1].
        '''
        T = self.T(y)
        n = np.sum(weights)
        return [self.tau[j] + np.sum(weights*T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n]    
    
    def update(self, y, weights = None):
        '''
        Update hyperparameters (tau) with data (y).
        
        Parameters
        ----------
        y: array
            Data.
        weights: array, optional
            Vector of weights for each data point. Each should be in the range [0, 1].
            Defaults to None.
        '''
        if weights is None:
            self.tau = self.tau_prime(y)
        else:
            self.tau = self.weighted_tau_prime(y, weights)
        self.hpar = self.tau_to_hpar(self.tau)
    
    def log_density(self, theta):
        '''
        The log-density of parameters (theta) given current hyperparameters (tau).
        
        Parameters
        ----------
        theta: list
            Distribution parameters (in natural form).
        '''
        return np.sum([np.sum(self.tau[j]*theta[j]) for j in range(self.p)]) - self.tau[self.p]*self.g(theta) - self.h(self.tau)
        
    def density(self, theta):
        '''
        The density of parameters (theta) given current hyperparameters (tau).
        
        Parameters
        ----------
        theta: list
            Distribution parameters (in natural form).
        '''
        return np.exp(self.log_density(theta))
    
    def log_marginal(self, y):
        '''
        The log of the marginal likelihood of data.
        
        Parameters
        ----------
        y: array
            Data.
        '''
        return -np.sum(self.f(y)) - self.h(self.tau, False) + self.h(self.tau_prime(y), False)            
        
    def weighted_log_marginal(self, y, weights):
        '''
        The log-marginal with each data point given a weight.
        
        Parameters
        ----------
        y: array
            Data.
        weights: array
            Vector of weights for each data point. Each should be in the range [0, 1].
        '''
        return -np.sum(weights*self.f(y)) - self.h(self.tau, False) + self.h(self.weighted_tau_prime(y, weights), False)
    
    def frac_log_marginal(self, y, frac):
        '''
        The log-marginal with each data point given the same weight.
        Used in computing fractional Bayes factors.
        
        Parameters
        ----------
        y: array
            Data.
        frac: numeric
            Weighting fraction used for all data points. Should be in the range [0, 1].
        '''
        n = y.shape[-1]
        return self.weighted_log_marginal(y = y, weights = frac*np.ones(n))
    
    def marginal(self, y):
        '''
        The marginal likelihood of data.
        
        Parameters
        ----------
        y: array
            Data.
        '''
        return np.exp(self.log_marginal(y))
    
    def entropy(self):
        '''
        The entropy of the distribution over theta.
        
        Notes
        -----
        .. math::
        \text{entropy} = -E[\log p(\theta)] = -\sum_{j=1}^p \ip{\tau_j, E[\theta_j]} + \tau_{p+1} E[g(\theta)] + h(\tau)
        '''
        return -np.sum([np.sum(self.tau[j]*self.E_theta()[j]) for j in range(self.p)]) + self.tau[self.p]*self.E_g_theta() + self.h(self.tau)
        

class multi_group_distribution:
    '''
    Class for a collection exponential family distributions of the same type with different
    parameters for different groups of observations.
    
    Notes
    -----
    This assumes no shared parameter (e.g. variance/precision) across groups.
    
    z gives the 1-hot encoding of each observation's group (row = group, column = observation).
    
    I should revise __init__ in the future to allow the groups to have different starting hyperparameters.
    '''
    
    def __init__(self, base_dist, hpar, n_t):
        '''
        Parameters
        ----------
        base_dist: distribution
            The distribution class.
        hpar: dict
            Initial hyperparameters (the same for each group).
            Specified in conventional form.
        n_t: integer
            Number of groups.
        '''
        self.n_t = n_t
        self.dists = []
        for t in range(self.n_t):
            self.dists += [base_dist(hpar)]
        self.tau = [self.dists[t].tau for t in range(self.n_t)]
        self.hpar = [self.dists[t].hpar for t in range(self.n_t)]
        
    def hpar_table(self):
        '''
        Put all of the (conventional) hyperparameters in a table (Pandas data frame).
        '''
        hpar_names = list(self.dists[0].hpar.keys())
        table = pd.DataFrame(0.0, index = range(self.n_t), columns = hpar_names)
        for t in range(self.n_t):
            for hpar_name in hpar_names:
                table.loc[t, hpar_name] = self.dists[t].hpar[hpar_name]
        return table
        
    def update(self, y, z):
        '''
        Update hyperparameters (tau) with data (y).
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            1-hot encoding of group membership.
        '''
        for t in range(self.n_t):
            self.dists[t].update(y, weights = z[t,:])
            self.tau[t] = self.dists[t].tau
            self.hpar[t] = self.dists[t].hpar
            
    def log_density(self, theta):
        '''
        The log-density of parameters (theta) given current hyperparameters (tau).
        
        Parameters
        ----------
        theta: list of lists
            Each i-th element of the list is another list representing the
            parameters (in natural form) of the i-th group's distribution.
        '''
        return np.sum([self.dists[t].log_density(theta[t]) for t in range(self.n_t)])
    
    def density(self, theta):
        '''
        The density of parameters (theta) given current hyperparameters (tau).
        
        Parameters
        ----------
        theta: list of lists
            Each i-th element of the list is another list representing the
            parameters (in natural form) of the i-th group's distribution.
        '''
        return np.exp(self.log_density(theta))
    
    def log_marginal(self, y, z):
        '''
        The log of the marginal likelihood of data.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            1-hot encoding of group membership.
        '''
        z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
        return np.sum([self.dists[t].log_marginal(y[z_flat == t]) for t in range(self.n_t)])
    
    def frac_log_marginal(self, y, z, frac):
        '''
        The log-marginal with each data point given the same weight.
        Used in computing fractional Bayes factors.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            1-hot encoding of group membership.
        frac: numeric
            Weighting fraction used for all data points. Should be in the range [0, 1].
        '''
        return np.sum([self.dists[t].weighted_log_marginal(y, weights = frac[t]*z[t,:]) for t in range(self.n_t)])
    
    def marginal(self, y, z):
        '''
        The marginal likelihood of data.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            1-hot encoding of group membership.
        '''
        return np.exp(self.log_marginal(y, z))
    
class distribution_with_predictor:
    '''
    Class for exponential familiy distributions with a predictor variable.
    
    Notes
    -----
    It is assumed that the last axis in T, z, S(z) etc. represents individual observations, e.g. participants.
    Thus, np.sum(T[j], axis = -1) takes the sum across observations.
    '''
        
    def tau_prime(self, y, z):
        '''
        Hyperparameters (tau) plus sufficient statistics of data (T(y, z)).
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        '''
        T = self.T(y, z)
        n = y.shape[-1]
        return [self.tau[j] + np.sum(T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n, self.tau[self.p + 1] + np.sum(self.S(z), axis = -1)]
    
    def weighted_tau_prime(self, y, z, weights):
        '''
        Hyperparameters (tau) plus sufficient statistics of data (T(y, z)), where each observation
        has a weight.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        weights: array
            Vector of weights for each data point. Each should be in the range [0, 1].
        '''
        T = self.T(y, z)
        n = np.sum(weights)
        return [self.tau[j] + np.sum(weights*T[j], axis = -1) for j in range(self.p)] + [self.tau[self.p] + n, self.tau[self.p + 1] + np.sum(weights*self.S(z), axis = -1)] 
    
    def update(self, y, z):
        '''
        Update hyperparameters (tau) with data (y) and predictor variables (z).
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        '''
        self.tau = self.tau_prime(y, z)
        self.hpar = self.tau_to_hpar(self.tau)
    
    def log_density(self, theta):
        '''
        The log-density of parameters (theta) given current hyperparameters (tau).
        
        Parameters
        ----------
        theta: list
            Distribution parameters (in natural form).
        '''
        return np.sum([np.sum(self.tau[j]*theta[j]) for j in range(self.p)]) - self.tau[self.p]*self.g(theta) - np.sum(self.tau[self.p + 1]*self.k(theta)) - self.h(self.tau)
        
    def density(self, theta):
        '''
        The density of parameters (theta) given current hyperparameters (tau).
        
        Parameters
        ----------
        theta: list
            Distribution parameters (in natural form).
        '''
        return np.exp(self.log_density(theta))
    
    def log_marginal(self, y, z):
        '''
        The log of the marginal likelihood of data.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        '''
        return -np.sum(self.f(y)) - self.h(self.tau, False) + self.h(self.tau_prime(y, z), False)
    
    def weighted_log_marginal(self, y, z, weights):
        '''
        The log-marginal with each data point given a weight.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        weights: array
            Vector of weights for each data point. Each should be in the range [0, 1].
        '''
        return -np.sum(weights*self.f(y)) - self.h(self.tau, False) + self.h(self.weighted_tau_prime(y, z, weights), False)
    
    def frac_log_marginal(self, y, z, frac):
        '''
        The log-marginal with each data point given the same weight.
        Used in computing fractional Bayes factors.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        frac: numeric
            Weighting fraction used for all data points. Should be in the range [0, 1].
        '''
        n = y.shape[-1]
        return self.weighted_log_marginal(y = y, z = z, weights = frac*np.ones(n))
    
    def marginal(self, y, z):
        '''
        The marginal likelihood of data.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        '''
        return np.exp(self.log_marginal(y, z))   
    
class beta_bernoulli(distribution):
    
    def __init__(self, hpar = {'a': 1.0, 'b': 1.0}):
        '''
        Initialize the distribution with given hyperparameters.
        
        Parameters
        ----------
        hpar: dict
            Initial (conventional) hyperparameters.
        '''
        self.hpar = hpar # conventional hyperparameters
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 1 # number of likelihood parameters
    
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['a'] - 1, hpar['a'] + hpar['b'] - 2]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'a': tau[1-1] + 1, 'b': tau[2-1] - tau[1-1] + 1}
    
    def par_to_theta(self, par):
        '''
        Convert conventional parameters to natural ones (theta).
        
        Parameters
        ----------
        par: dict
            Conventional parameters. Should contain 'psi'.
        '''
        return [np.log(par['psi']/(1 - par['psi']))]
    
    def T(self, y):
        '''
        Sufficient statistics.
        
        Parameters
        ----------
        y: array
            Data.
        '''
        return [y]
        
    def f(self, y):
        return 0.0*y
        
    def g(self, theta):
        return np.log(1 + np.exp(theta))
        
    def h(self, tau, include_constant = True):
        '''
        Parameters
        ----------
        tau: list
            Distribution hyperparameters (in natural form).
        include_constant: logical, optional
            Whether to include constant terms, i.e. terms that do not
            depend on tau. Defaults to True.
            
        Notes
        -----
        Omitting constant terms improves the accuracy of calculating the marginal likelihood.
        This calculation involves the subtraction :math:`h(\tau) - h(\tau')`. The constant terms in h(tau)
        and h(tau') cancel out, so mathematically it's fine to omit them. Including them tends
        to cause more catastrophic cancellation in the floating point arithmetic.
        
        In this particular distribution there aren't actually any constant terms in h, but we retain
        the include_constant parameter and explanation for the sake of consistency.
        '''
        hpar = self.tau_to_hpar(tau)
        return betaln(hpar['a'], hpar['b'])

class normal_known_precision(distribution):
    '''
    Normal likelihood with known precision (xi) and a normal
    prior on the mean (mu).
    
    DOUBLE CHECK THIS (SOMETHING SEEMS TO BE WRONG).
    '''
    def __init__(self, hpar = {'m_mu': 0.0, 'xi_mu': 1.0}, xi = 1.0):
        '''
        Initialize the distribution with given hyperparameters.
        
        Parameters
        ----------
        hpar: dict
            Initial (conventional) hyperparameters.
        '''
        self.xi = xi
        self.sigma = 1/np.sqrt(self.xi)
        self.hpar = hpar
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 1
        
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [self.hpar['m_mu']*self.sigma*self.hpar['xi_mu'], self.hpar['xi_mu']/self.xi]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'m_mu': self.sigma*self.tau[1-1]/self.tau[2-1], 'xi_mu': self.xi*self.tau[2-1]}
    
    def par_to_theta(self, par):
        '''
        Convert conventional parameters to natural ones (theta).
        
        Parameters
        ----------
        par: dict
            Conventional parameters. Should contain 'mu'.
        '''
        return np.array(par['mu']/self.sigma)
    
    def T(self, y):
        '''
        Sufficient statistics.
        
        Parameters
        ----------
        y: array
            Data.
        '''
        return [y/self.sigma]
    
    def f(self, y):
        return 0.5*(self.xi*y**2 + np.log(2*np.pi) - np.log(self.xi))
    
    def g(self, theta):
        return 0.5*theta**2
        
    def h(self, tau, include_constant = True):
        '''
        Parameters
        ----------
        tau: list
            Distribution hyperparameters (in natural form).
        include_constant: logical, optional
            Whether to include constant terms, i.e. terms that do not
            depend on tau. Defaults to True.
            
        Notes
        -----
        Omitting constant terms improves the accuracy of calculating the marginal likelihood.
        This calculation involves the subtraction :math:`h(\tau) - h(\tau')`. The constant terms in h(tau)
        and h(tau') cancel out, so mathematically it's fine to omit them. Including them tends
        to cause more catastrophic cancellation in the floating point arithmetic.
        '''
        hpar = self.tau_to_hpar(tau)
        h = 0.5*(hpar['xi_mu']*hpar['m_mu']**2 - np.log(hpar['xi_mu']) + np.log(self.xi))
        if include_constant:
            h += 0.5*np.log(2*np.pi)
        return h
    
class normal_gamma_normal(distribution):
    '''
    Univariate normal likelihood with a normal-gamma prior.
    
    Notes
    -----
    ** ADD MATH **
    '''
    
    def __init__(self, hpar = {'m': 0.0, 'lambda': 1.0, 'alpha': 1.0, 'beta': 1.0}):
        '''
        Initialize the distribution with given hyperparameters.
        
        Parameters
        ----------
        hpar: dict
            Initial (conventional) hyperparameters.
        '''
        self.hpar = hpar
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 2 # number of parameters
    
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['lambda']*hpar['m'],
                -hpar['beta'] - 0.5*hpar['lambda']*hpar['m']**2,
                2*hpar['alpha'] - 1]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'m': tau[1-1]/tau[3-1],
                'lambda': tau[3-1],
                'alpha': 0.5*(tau[3-1] + 1),
                'beta': -0.5*tau[1-1]**2/tau[3-1] - tau[2-1]}
    
    def par_to_theta(self, par):
        '''
        Convert conventional parameters to natural ones (theta).
        
        Parameters
        ----------
        par: dict
            Conventional parameters. Should contain 'mu' and 'sigma'.
        '''
        return [par['mu']/par['sigma']**2, 
                1/par['sigma']**2]
    
    def T(self, y):
        '''
        Sufficient statistics.
        
        Parameters
        ----------
        y: array
            Data.
        '''
        return [y, -0.5*y**2]      
        
    def f(self, y):
        n = y.shape[-1]
        return np.array(n*[0.5*np.log(2*np.pi)])
        
    def g(self, theta):
        return 0.5*(theta[1-1]**2/theta[2-1] - np.log(theta[2-1]))
        
    def h(self, tau, include_constant = True):
        '''
        Parameters
        ----------
        tau: list
            Distribution hyperparameters (in natural form).
        include_constant: logical, optional
            Whether to include constant terms, i.e. terms that do not
            depend on tau. Defaults to True.
            
        Notes
        -----
        Omitting constant terms improves the accuracy of calculating the marginal likelihood.
        This calculation involves the subtraction h(tau) - h(tau'). The constant terms in h(tau)
        and h(tau') cancel out, so mathematically it's fine to omit them. Including them tends
        to cause more catastrophic cancellation in the floating point arithmetic.
        '''
        hpar = self.tau_to_hpar(tau)
        h = gammaln(hpar['alpha']) - hpar['alpha']*np.log(hpar['beta']) - 0.5*np.log(hpar['lambda'])
        if include_constant:
            h += 0.5*np.log(2*np.pi)
        return h
    
    def E_theta(self):
        '''
        Compute :math:`E[\theta]`
        '''
        return [] # FINISH
        
    def E_g_theta(self):
        '''
        Compute :math:`E[g(\theta)]`
        '''
        return [] # FINISH

class mvnormal_Wishart_mvnormal(distribution):
    '''
    Multivariate normal likelihood with a multivariate normal-Wishart prior.
    
    ** DOUBLE CHECK EVERYTHING **
    
    Notes
    -----
    
    ** ADD MATH **
    '''
    
    def __init__(self, hpar):
        '''
        Initialize the distribution with given hyperparameters.
        
        Parameters
        ----------
        hpar: dict
            Initial (conventional) hyperparameters, consisting of:
            m: 1-d array
                Prior mean vector for mu.
            lambda: int
                Prior scaling parameter (see the class documentation).
            V_inv: 2-d square array
                Inverse of the prior scale matrix for the Wishart prior on Sigma.
        '''
        self.hpar = hpar
        self.tau = self.hpar_to_tau(self.hpar)
        self.k = hpar['m'].shape[0] # size of mu
        self.p = 2 # number of parameters
        
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['lambda']*hpar['m'],
                -0.5*(hpar['lambda']*np.outer(hpar['m'], hpar['m']) + hpar['V_inv']),
                hpar['lambda']]
        
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'m': tau[1-1]/tau[3-1],
                'V_inv': -2*tau[2-1] - (1/tau[3-1])*np.outer(tau[1-1], tau[1-1]),
                'lambda': tau[3-1]}
        
    def par_to_theta(self, par):
        '''
        Convert conventional parameters to natural ones (theta).
        
        Parameters
        ----------
        par: dict
            Conventional parameters. Should contain 'mu' and 'Sigma_inv'.
        '''
        return [par['Sigma_inv']@par['mu'], par['Sigma_inv']]
        
    def T(self, y):
        '''
        Sufficient statistics.
        
        Parameters
        ----------
        y: array
            Data.
        '''
        n = y.shape[-1]
        return [y, -0.5*np.stack([np.outer(y[:,i], y[:,i]) for i in range(n)], axis = -1)]
    
    def f(self, y):
        return 0.5*self.k*np.log(2*np.pi)
        
    def g(self, theta):
        return 0.5*( theta[1-1].T@inv(theta[2-1])@theta[1-1] - np.log(det(theta[2-1])) )
        
    def h(self, tau, include_constant = True):
        '''
        Parameters
        ----------
        tau: list
            Distribution hyperparameters (in natural form).
        include_constant: logical, optional
            Whether to include constant terms, i.e. terms that do not
            depend on tau. Defaults to True.
        '''
        h = -0.5*self.k*np.log(tau[3-1]) + 0.5*(tau[3-1] + self.k)*self.k*np.log(2) - 0.5*(tau[3-1] + self.k)*np.log(det( -2*tau[2-1] - (1/tau[3-1])*np.outer(tau[1-1], tau[1-1]) )) + multigammaln(0.5*(tau[3-1] + self.k), self.k)
        if include_constant:
            h += 0.5*self.k*np.log(2*np.pi)
        return h
    
class multi_group_normal(distribution_with_predictor):
    '''
    Notes
    -----
    z is an n_t x n array indicating group membership.
    z[t, i] = 1 if observation i is in group t and = 0 otherwise.
    
    theta = [xi*mu, xi]
    mu = the vector of the means for y in each group
    xi = the precision of y (shared across groups)
    '''
    
    def __init__(self, hpar = {'m': 0.0, 'lambda': 1.0, 'alpha': 1.0, 'beta': 1.0}, n_t = 2):
        '''
        Parameters
        ----------
        hpar: dict
            Initial (conventional) hyperparameters.
        n_t: int
            Number of groups.
        '''
        self.n_t = n_t
        self.hpar = {'m': hpar['m']*np.ones(n_t), 
                     'lambda': hpar['lambda']*np.ones(n_t), 
                     'alpha': hpar['alpha'], 
                     'beta': hpar['alpha']}
        self.tau = self.hpar_to_tau(self.hpar)
        self.p = 2
    
    def hpar_to_tau(self, hpar):
        '''
        Convert conventional hyperparameters to natural ones (tau).
        '''
        return [hpar['lambda']*hpar['m'], 
                -hpar['beta'] - 0.5*np.sum(hpar['lambda']*hpar['m']**2),
                2*hpar['alpha'] + self.n_t - 2,
                0.5*hpar['lambda']]
    
    def tau_to_hpar(self, tau):
        '''
        Convert natural hyperparameters (tau) to conventional ones.
        '''
        return {'m': 0.5*tau[1-1]/tau[4-1],
                'lambda': 2*tau[4-1],
                'alpha': 0.5*(tau[3-1] - self.n_t) + 1,
                'beta': -tau[2-1] - 0.25*np.sum(tau[1-1]**2/tau[4-1])}
    
    def T(self, y, z):
        '''
        Sufficient statistics.
        
        Parameters
        ----------
        y: array
            Data.
        z: array
            Predictor variables.
        '''
        return [z*y, -0.5*y**2]
    
    def S(self, z):
        return 0.5*z
    
    def f(self, y):
        n = y.shape[-1]
        return np.array(n*[0.5*np.log(2*np.pi)])
        
    def g(self, theta):
        return -0.5*np.log(theta[2-1])
        
    def k(self, theta):
        return theta[1-1]**2/theta[2-1]
        
    def h(self, tau, include_constant = True):
        '''
        Parameters
        ----------
        tau: list
            Distribution hyperparameters (in natural form).
        include_constant: logical, optional
            Whether to include constant terms, i.e. terms that do not
            depend on tau. Defaults to True.
            
        Notes
        -----
        Omitting constant terms improves the accuracy of calculating the marginal likelihood.
        This calculation involves the subtraction h(tau) - h(tau'). The constant terms in h(tau)
        and h(tau') cancel out, so mathematically it's fine to omit them. Including them tends
        to cause more catastrophic cancellation in the floating point arithmetic.
        '''
        hpar = self.tau_to_hpar(tau)
        h = gammaln(hpar['alpha']) - hpar['alpha']*np.log(hpar['beta']) - 0.5*np.sum(np.log(hpar['lambda']))
        if include_constant:
            h += 0.5*self.n_t*np.log(2*np.pi)
        return h
        
    


import numpy as np
import pandas as pd
import functools
import multiprocessing
from time import perf_counter
from .lpa import lpa_model
from scipy.cluster.vq import kmeans2

def fit_mixture_model(x,
                      model_class = lpa_model,
                      n_workers = 4,
                      kmeans = False,
                      restarts = 20, 
                      T = 20,
                      seed = 1234, 
                      tolerance = 1e-05, 
                      max_iter = 1000, 
                      min_iter = 10,
                      **prior_hpar):
    """
    Fit a mixture model using mean field variational Bayes.
    
    Parameters
    ----------
    x: data frame or array
        Observed data (data frame or matrix).  If a matrix then rows are variables and columns
        are observations; if a data frame then rows are observations and columns are variables.
    model_class: class, optional
        Class of mixture model to fit.
    n_workers: int, optional
        Number of workers in the pool for parallel processing; should be less than
        or equal to the number of available CPUs.
    kmeans: logical, optional
        Should the k-means++ algorithm be used for initial starting points?
        If so, then k-means clusterings of different sizes are produced and used.
    restarts: int, optional
        Number of random restarts.
    T: integer, optional
        Truncation level of the variational approximation.
    seed: int, optional
        Random seed (determines starting conditions of each optimization).
    tolerance: float, optional
        Relative change in the ELBO at which the optimization should stop.
    max_iter: integer, optional
        Maximum number of iterations to run the optimization.
    min_iter: integer, optional
        Minimum number of iterations to run the optimization.
    prior_hpar:
        Prior hyperparameters (specified with keywords).
        
    Notes
    -----
    One fits a statistical model using variational Bayes by maximizing a quantity called
    the ELBO (evidence lower bound). The best variational Bayes approximation of the true
    model posterior distribution is the one with the highest ELBO. The algorithm for maximizing
    the ELBO unfortunately only finds local maxima, so to have a decent chance of finding the
    global maximum (truly best approximation to the real model posterior distribution) we need
    different starting points of the variational model hyperparameters. This function provides two
    ways to choose starting points. First, the k-means++ algorithm (a simple clustering algorithm)
    can be used to produce initial clustering of participants. Second, participants can be randomly
    assigned (in a 'soft' way) to different mixture components/latent classes/latent profiles. These
    two methods are not exclusive: both types of starting point will be used if 'kmeans' is set to
    True and 'restarts' > 0. The function creates a separate mixture model object with each starting
    point and fits them using parallel processing for speed.
    
    Output
    ------
    A dictionary with the following:
    
    final_elbo: list
        Final ELBO (evidence lower bound) values of each model.
    model_list: list
        Fitted model objects.
    best_model: mixture_model object
        Best fitted model (i.e. the one with the highest ELBO).
    fit_time: float
        Total time used to fit.
    """
    tic = perf_counter()

    model_list = []
    final_elbo = []
    
    # run using starting points initialized with k-means
    if kmeans:
        if isinstance(x, pd.DataFrame):
            x = x.values.copy().transpose()
        n = x.shape[1]
        phi_list = []
        for Tval in range(2, T + 1, 3):
            centroid, label = kmeans2(x.T, k = Tval, minit = '++', seed = seed)
            #phi = np.zeros([Tval, n])
            phi = np.ones([Tval, n])*0.3/(Tval - 1)
            for i in range(n):
                #phi[label[i], i] = 1.0
                phi[label[i], i] = 0.7
            phi_list += [phi]
            
        with multiprocessing.Pool(n_workers) as pool:
            fun = functools.partial(_model_fit_wrapper_phi, x = x, T = T, model_class = model_class, tolerance = tolerance, max_iter = max_iter, min_iter = min_iter, **prior_hpar)        
            results = pool.map(fun, phi_list)
            for result in results:
                model_list += [result]
                final_elbo += [result.elbo_list[-1]]
    
    # run using random starting points
    if restarts > 0:
        with multiprocessing.Pool(n_workers) as pool:
            fun = functools.partial(_model_fit_wrapper_seed, x = x, T = T, model_class = model_class, tolerance = tolerance, max_iter = max_iter, min_iter = min_iter, **prior_hpar)

            # generate random seeds for each model based on the seed parameter
            seed_list = []
            rng = np.random.default_rng(seed)
            for i in range(restarts):
                seed_list += [rng.integers(low = 1000, high = 9999)]

            results = pool.map(fun, seed_list)
            for result in results:
                model_list += [result]
                final_elbo += [result.elbo_list[-1]]
        
    final_elbo = np.array(final_elbo)
    toc = perf_counter()
    
    return {'final_elbo': final_elbo, 'model_list': model_list, 'best_model': model_list[final_elbo.argmax()], 'fit_time': toc - tic}

def _model_fit_wrapper_seed(seed, x, T, model_class, tolerance, max_iter, min_iter, **prior_hpar):
    """
    Defines a mixture model and fits it to the data with specified random seed, returning the result.
    This function is only defined in order to get the parallelization to work.
    """
    new_model = model_class(x = x, T = T, seed = seed, **prior_hpar)
    new_model.fit(tolerance = tolerance, max_iter = max_iter, min_iter = min_iter)
    return new_model

def _model_fit_wrapper_phi(phi, x, T, model_class, tolerance, max_iter, min_iter, **prior_hpar):
    """
    Defines an LPA model and fits it to the data with specified starting value of phi, returning the result.
    This function is only defined in order to get the parallelization to work.
    """
    new_model = model_class(x = x, phi = phi, T = T, seed = None, **prior_hpar)
    new_model.fit(tolerance = tolerance, max_iter = max_iter, min_iter = min_iter)
    return new_model


import numpy as np
import pandas as pd
from numpy.testing import *
from scipy.stats import beta as beta_dist
from scipy.stats import gamma as gamma_dist
from scipy.stats import norm as norm_dist
from scipy.stats import multivariate_normal, multivariate_t, wishart
from scipy.stats import t as t_dist
from scipy.linalg import inv
from vbayesfa import exp_families as ef

# to run all tests: pytest /Users/sampaskewitz/Documents/vbayesfa/src/vbayesfa/tests

def _assert_dict_almost_equal(dict1, dict2):
    # This is just a convenience function to test equality of the elements of two dicts.
    for key in dict1.keys():
        assert_almost_equal(dict1[key], dict2[key])

def test_normal():
    # define the distribution
    dist = ef.normal_gamma_normal()
    
    # test updating
    y = norm_dist.rvs(size = 4, random_state = 1234) # some data
    dist.update(y)
    assert_almost_equal(list(dist.hpar.values()), # updated hyperparameters from my code
                        [0.0801029082720826, 5.0, 3.0, 2.8794961524616207]) # correct hyperparameters

    # test parameter distribution
    test_par = {'mu': 1.2, 'sigma': 0.39} # some values to test
    scipy_result = norm_dist.logpdf(test_par['mu'], dist.hpar['m'], test_par['sigma']/np.sqrt(dist.hpar['lambda'])) + gamma_dist.logpdf(1/test_par['sigma']**2, a = dist.hpar['alpha'], scale = 1/dist.hpar['beta']) # result from scipy.stats
    our_result = dist.log_density(dist.par_to_theta(test_par))
    assert_almost_equal(our_result, 
                        scipy_result)
    
    # test marginal distribution
    y_new = np.array([-9.0]) # some data value to test
    scipy_log_marginal = t_dist.logpdf(y_new, 
                                       df = 2*dist.hpar['alpha'],
                                       loc = dist.hpar['m'],
                                       scale = np.sqrt(dist.hpar['beta']*(dist.hpar['lambda'] + 1)/(dist.hpar['lambda']*dist.hpar['alpha'])))
    assert_almost_equal(dist.log_marginal(y_new), # result from my code
                        scipy_log_marginal) # result from scipy.stats

    
def test_bernoulli():
    # define the distribution
    dist = ef.beta_bernoulli()
    
    # test updating
    y = np.array([1, 0, 0, 1, 1, 0, 1]) # some data
    dist.update(y)
    assert_almost_equal(list(dist.hpar.values()), # updated hyperparameters from my code
                        [5.0, 4.0]) # correct hyperparameters

    # test parameter distribution
    test_par = {'psi': 0.4} # some value of the parameter to test
    assert_almost_equal(dist.log_density(dist.par_to_theta(test_par)), # result from my code
                        beta_dist.logpdf(test_par['psi'], dist.hpar['a'], dist.hpar['b'])) # result from scipy.stats
    
    # test marginal distribution
    y_new = 1 # some data value to test
    prob = dist.hpar['a']/(dist.hpar['a'] + dist.hpar['b']) # marginal probability that y_new = 1
    assert_almost_equal(dist.log_marginal(np.array([y_new])), # result from my code
                        y_new*np.log(prob) + (1 - y_new)*np.log(1 - prob)) # correct result
    
def test_mvnormal():
    # define the distribution
    dist = ef.mvnormal_Wishart_mvnormal({'m': np.zeros(2),
                                         'lambda': 2,
                                         'V_inv': np.eye(2)})
    
    # test hyperparameter conversion (convert hpar to tau then back to hpar and check that it's the same)
    test_hpar = {'m': np.array([0.3, -1.2]), 
                 'V_inv': inv(np.array([2.0, 0.1, 0.1, 2.0]).reshape([2, 2])), 
                 'lambda': 9}
    _assert_dict_almost_equal(dist.tau_to_hpar(dist.hpar_to_tau(test_hpar)),
                              test_hpar)
    
    # test updating
    n = 5
    y = multivariate_normal.rvs(size = n, cov = inv(np.array([1.0, 0.3, 0.3, 1.0]).reshape([2, 2])), random_state = 1234).T # some data (transposed so that observations are indexed by the last axis)
    ybar = y.mean(axis = -1)
    comparison_updated_hpar = {'m': (dist.hpar['lambda']*dist.hpar['m'] + n*ybar)/(dist.hpar['lambda'] + n),
                              'V_inv': dist.hpar['V_inv'] + np.sum(np.stack([np.outer(y[:,i] - ybar, y[:,i] - ybar) for i in range(n)], axis = -1), axis = -1) + ((dist.hpar['lambda']*n)/(dist.hpar['lambda'] + n))*np.outer(ybar - dist.hpar['m'], ybar - dist.hpar['m']), 
                              'lambda': dist.hpar['lambda'] + n}
    dist.update(y)
    _assert_dict_almost_equal(dist.hpar, # updated hyperparameters from my code
                              comparison_updated_hpar) # correct hyperparameters
    
    # test parameter distribution
    test_par = {'mu': np.array([1.2, -2.0]),
                'Sigma_inv': inv(np.array([2.3, -0.9, -0.9, 2.3]).reshape([2, 2]))} # some values to test
    scipy_result = multivariate_normal.logpdf(test_par['mu'], dist.hpar['m'], inv(test_par['Sigma_inv'])/dist.hpar['lambda']) + wishart.logpdf(test_par['Sigma_inv'], dist.hpar['lambda'] + dist.k, inv(dist.hpar['V_inv']))
    assert_almost_equal(dist.log_density(dist.par_to_theta(test_par)), # our result
                        scipy_result)
    
    # test marginal distribution
    y_new = multivariate_normal.rvs(size = 1, cov = inv(np.array([1.0, 0.3, 0.3, 1.0]).reshape([2, 2])), random_state = 9876).reshape([2, 1]) # some data value to test
    scipy_result = multivariate_t.logpdf(y_new.squeeze(),
                                         loc = dist.hpar['m'], 
                                         shape = dist.hpar['V_inv']/dist.hpar['lambda'],
                                         df = dist.hpar['lambda'] + 1) # from the conjugate prior article, with a little extra algebra
    assert_almost_equal(dist.log_marginal(y_new), # our result
                        scipy_result)
    


import numpy as np
import pandas as pd
from scipy.special import gamma as Gamma
from scipy.integrate import quad
from scipy.linalg import det, inv
from scipy.stats import invgamma
from . import exp_families

def _Helmert_contrast_matrix(p):
    '''
    Create a matrix of Helmert contrast codes for a 1-way ANOVA.
    
    Parameters
    ----------
    p: integer
        Number of factor levels/groups.
    '''
    Q_transpose = np.zeros([p-1, p])
    for i in range(p-1):
        new_row = np.array(i*[0] + [1] + (p-1-i)*[-1/(p-1-i)])
        new_row = new_row/np.sqrt(np.sum(new_row**2))
        Q_transpose[i,:] = new_row
    return Q_transpose.T

def _design_matrix(groups):
    '''
    Turn a "flat" representation of groups in a 1-way ANOVA into a design matrix with dummy variables.
    
    Parameters
    ----------
    groups: array-like
       "Flat" representation of groups/factor levels.
    '''
    n = groups.shape[0] # number of observations
    group_names = np.unique(groups) # names of factor levels
    p = group_names.shape[0] # number of factor levels (does not include the intercept)
    X = np.zeros([n, p])
    for j in range(p):
        X[groups == group_names[j], j] = 1.0
    return X

def _partition(collection):
    '''
    Generate the partitions of a collection.
    
    Parameters
    ----------
    collection: list or similar
        Collection of things to partition.
    
    This is based on the following: https://stackoverflow.com/questions/19368375/set-partitions-in-python
    '''
    if len(collection) == 1:
        yield [ collection ]
        return

    first = collection[0]
    for smaller in _partition(collection[1:]):
        # insert `first` in each of the subpartition's subsets
        for n, subset in enumerate(smaller):
            yield smaller[:n] + [[ first ] + subset]  + smaller[n+1:]
        # put `first` in its own subset 
        yield [ [ first ] ] + smaller
        
def _remove_nan_groups(y, groups):
    '''
    Convenience function to remove observations with missing data.
    '''
    is_nan = np.isnan(y)
    if np.sum(is_nan) > 0:
        y = y[~is_nan].copy()
        groups = groups[~is_nan].copy()
    return (y, groups)

def _remove_nan_1hot(y, z):
    '''
    Convenience function to remove observations with missing data.
    '''
    is_nan = np.isnan(y)
    if np.sum(is_nan) > 0:
        y = y[~is_nan].copy()
        z = z[:, ~is_nan].copy()
    return (y, z)

def bf10_1way_anova(y, groups, rscale = 0.5, epsrel = 1e-4, limit = 50):
    '''
    Bayes factor vs. the null model for the 1-way fixed effects ANOVA model,
    using JZS priors (not conjugate ones).
    See Equations 8, 13 etc. in the Rouder et al paper.
    
    Parameters
    ----------
    y: array
        Dependent variable array.
    groups: array
        Array (vector) of groups (specified e.g. with character strings).
    rscale: float
        Specifies the width of the inverse gamma distribution on fixed effects.
    epsrel: float, optional
        Relative error tolerance for numerical integration.
    limit: int, optional
        An upper bound on the number of subintervals used in the adaptive algorithm
        for numerical integration.
    
    Returns
    ------
    A tuple containing the Bayes factor (element 0) and numerical integration 
    error (element 1).
    
    Notes
    -----
    
    This uses the following prior distribution:
    
    g ~ InvGamma(0.5, 0.5*rscale^2)
    
    With this interpretation, rscale corresponds to the rscaleFixed parameter
    from the BayesFactor R package. Note that the documentation for that package
    incorrectly specifies that g ~ InvGamma(0.5, 0.5*rscale), i.e. that rscale
    is not squared. I have confirmed that rscale is in fact squared in their
    implementation, so I have also squared rscale here for consistency.
    
    The BayesFactor R package gives the following labels to different values of
    rscale:
    
    1/2 = "medium"
    sqrt(2)/2 = "wide"
    1 = "ultrawide"
    
    I have decided to implement the "medium" size as the default.
    '''
    # remove missing data, if any
    (y, groups) = _remove_nan_groups(y, groups)
    
    # make the design matrix (omitting the intercept term)
    X = _design_matrix(groups)
    n = X.shape[0] # number of participants/observations
            
    # project the design matrix down into a lower subspace
    p = X.shape[1] # number of factor levels
    Q = _Helmert_contrast_matrix(p)
    X_star = X@Q
    
    # define a wrapper function for T_m times the inverse gamma prior on g    
    P0 = np.ones([n, n])/n
    y_tilde = (np.eye(n) - P0)@y
    X_tilde = (np.eye(n) - P0)@X_star
    numerator = y.T@y - n*y.mean()**2
    y_tilde_inner = y_tilde.T@y_tilde
    def _integrand(g):
        G_inv = np.diag(1/np.array((p-1)*[g]))
        det_G = g**(p-1)
        Vg = X_tilde.T@X_tilde + G_inv
        Vg_inv = inv(Vg, check_finite = False)
        denominator = y_tilde_inner - y_tilde.T@X_tilde@Vg_inv@X_tilde.T@y_tilde
        term1 = (numerator/denominator)**(n-1)
        term2 = 1/(det_G*det(Vg, overwrite_a = True, check_finite = False))
        S = (term1*term2)**0.5
        return S*invgamma.pdf(g, a = 0.5, scale = 0.5*rscale**2)
    
    # numerically integrate across g to obtain the marginal likelihood
    return quad(_integrand, a = 0, b = np.inf, epsrel = epsrel, limit = limit)[0]

def partition_posthoc(y, groups, rscale = 0.5, epsrel = 1e-2, limit = 50):
    '''
    Find the best partition of groups/factor levels as a post-hoc analysis
    following a Bayesian 1-way ANOVA.
    
    Parameters
    ----------
    y: array
        Dependent variable array.
    groups: array
        Array (vector) of groups (specified e.g. with character strings).
    rscale: float
        Specifies the width of the inverse gamma distribution on fixed effects.
    epsrel: float, optional
        Relative error tolerance for numerical integration.
    limit: int, optional
        An upper bound on the number of subintervals used in the adaptive algorithm
        for numerical integration.
    
    Returns
    -------
    A dictionary with the following elements:
    
    best_partition:
        The best partition (the one with the highest evidence).
    partition_list:
        List of all partitions.
    evidence:
        The evidence for each partition.
    
    Notes
    -----
    
    ** ADD THIS **
    '''
    # remove missing data, if any
    (y, groups) = _remove_nan_groups(y, groups)
    
    # count things etc.
    n = y.shape[0] # number of observations
    group_names = np.unique(groups) # names of factor levels
    p = group_names.shape[0] # number of factor levels (does not include the intercept)
    
    # list partitions
    partition_list = list(_partition(list(range(p)))) # list of all partitions
    partition_list.pop(0) # remove the initial partition, which has everyone in the same group (H0)
    n_partitions = len(partition_list) # number of partitions
    
    # compute the Bayes factor (vs. the null model) for each partition
    bf = np.zeros(n_partitions) # array for storing each partition's Bayes factor
    for i in range(n_partitions):        
        # create a mapper from the original groups to those based on the current partition        
        mapper = pd.Series(index = group_names)
        for j in range(len(partition_list[i])):
            mapper[partition_list[i][j]] = j
        
        # compute the Bayes factor (vs. the null model) for this partition
        bf[i] = bf10_1way_anova(y, mapper[groups].values, epsrel = epsrel, limit = limit)
        
    return {'best_partition': partition_list[bf.argmax()],
            'partition_list': partition_list,  
            'bf': bf}

def normal_shared_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    '''
    Bayes factor for equality of group means (ANOVA), assuming equal variance across groups.
    This uses a conjugate prior.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_m: float, optional
        Prior m hyperparameter.
    prior_lambda: float, optional
        Prior lambda hyperparameter.
    prior_alpha: float, optional
        Prior alpha hyperparameter.
    prior_beta: float, optional
        Prior beta hyperparameter.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_normal(hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def post_normal_shared_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    '''
    Posterior Bayes factor for equality of group means (ANOVA), assuming equal variance across groups.
    This uses a conjugate prior.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_m: float, optional
        Prior m hyperparameter.
    prior_lambda: float, optional
        Prior lambda hyperparameter.
    prior_alpha: float, optional
        Prior alpha hyperparameter.
    prior_beta: float, optional
        Prior beta hyperparameter.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_normal(hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    H1.update(y, z)
    H0.update(y)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def frac_normal_shared_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    '''
    Fractional Bayes factor for equality of group means (ANOVA), assuming equal variance across groups.
    This uses a conjugate prior.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_m: float, optional
        Prior m hyperparameter.
    prior_lambda: float, optional
        Prior lambda hyperparameter.
    prior_alpha: float, optional
        Prior alpha hyperparameter.
    prior_beta: float, optional
        Prior beta hyperparameter.
        
    Notes
    -----
    Each group has a different training fraction, determined as 1/sqrt(n), where n is the
    sample size of that group.
    '''
    (y, z) = _remove_nan_1hot(y, z)    
    n_per_group = z.sum(axis = -1)
    z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
    frac = 1/np.sqrt(n_per_group) # different training fractions for each group
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_normal(hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = (H1.log_marginal(y, z) - H1.weighted_log_marginal(y, z, weights = frac[z_flat])) - (H0.log_marginal(y) - H0.weighted_log_marginal(y, weights = frac[z_flat]))
    return np.exp(log_bf10)

def normal_dif_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    '''
    Bayes factor for equality of group means (ANOVA), assuming different variances across groups.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_m: float, optional
        Prior m hyperparameter.
    prior_lambda: float, optional
        Prior lambda hyperparameter.
    prior_alpha: float, optional
        Prior alpha hyperparameter.
    prior_beta: float, optional
        Prior beta hyperparameter.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_distribution(exp_families.normal_gamma_normal, hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def frac_normal_dif_var(y, z, prior_m = 0.0, prior_lambda = 1.0, prior_alpha = 1.0, prior_beta = 1.0):
    '''
    Fractional Bayes factor for equality of group means (ANOVA), assuming different variances across groups.
    This uses a conjugate prior.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_m: float, optional
        Prior m hyperparameter.
    prior_lambda: float, optional
        Prior lambda hyperparameter.
    prior_alpha: float, optional
        Prior alpha hyperparameter.
    prior_beta: float, optional
        Prior beta hyperparameter.
        
    Notes
    -----
    Each group has a different training fraction, determined as 1/sqrt(n), where n is the
    sample size of that group.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_per_group = z.sum(axis = -1)
    z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
    frac = 1/np.sqrt(n_per_group) # different training fractions for each group
    n_t = z.shape[0] # number of groups
    hpar = {'m': prior_m, 'lambda': prior_lambda, 'alpha': prior_alpha, 'beta': prior_beta} # prior hyperparameters
    H1 = exp_families.multi_group_distribution(exp_families.normal_gamma_normal, hpar, n_t)
    H0 = exp_families.normal_gamma_normal(hpar)
    log_bf10 = (H1.log_marginal(y, z) - H1.frac_log_marginal(y, z, frac = frac)) - (H0.log_marginal(y) - H0.weighted_log_marginal(y, weights = frac[z_flat]))
    return np.exp(log_bf10)

def bernoulli(y, z, prior_a = 0.5, prior_b = 0.5):
    '''
    Bayes factor for testing equality of outcome probabilities across groups.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_a: float, optional
        Prior a hyperparameter.
    prior_b: float, optional
        Prior b hyperparameter.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_t = z.shape[0] # number of groups
    H1 = exp_families.multi_group_distribution(exp_families.beta_bernoulli, 
                                               {'a': prior_a, 'b': prior_b}, 
                                               n_t)
    H0 = exp_families.beta_bernoulli({'a': prior_a, 'b': prior_b})
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def post_bernoulli(y, z, prior_a = 0.5, prior_b = 0.5):
    '''
    Posterior Bayes factor for testing equality of outcome probabilities across groups.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_a: float, optional
        Prior a hyperparameter.
    prior_b: float, optional
        Prior b hyperparameter.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_t = z.shape[0] # number of groups
    H1 = exp_families.multi_group_distribution(exp_families.beta_bernoulli, 
                                               {'a': prior_a, 'b': prior_b}, 
                                               n_t)
    H0 = exp_families.beta_bernoulli({'a': prior_a, 'b': prior_b})
    H1.update(y, z)
    H0.update(y)
    log_bf10 = H1.log_marginal(y, z) - H0.log_marginal(y)
    return np.exp(log_bf10)

def frac_bernoulli(y, z, prior_a = 0.5, prior_b = 0.5):
    '''
    Fractional Bayes factor for testing equality of outcome probabilities across groups.
    
    Parameters
    ----------
    y: array
        Data.
    z: array
        1-hot encoding of group membership.
    prior_a: float, optional
        Prior a hyperparameter.
    prior_b: float, optional
        Prior b hyperparameter.
        
    Notes
    -----
    Each group has a different training fraction, determined as 1/sqrt(n), where n is the
    sample size of that group.
    '''
    (y, z) = _remove_nan_1hot(y, z)
    n_per_group = z.sum(axis = -1)
    z_flat = z.argmax(axis = 0) # encodes group in a single vector (not 1-hot encoding)
    frac = 1/np.sqrt(n_per_group) # different training fractions for each group
    n_t = z.shape[0] # number of groups
    H1 = exp_families.multi_group_distribution(exp_families.beta_bernoulli, 
                                               {'a': prior_a, 'b': prior_b}, 
                                               n_t)
    H0 = exp_families.beta_bernoulli({'a': prior_a, 'b': prior_b})
    log_bf10 = (H1.log_marginal(y, z) - H1.frac_log_marginal(y, z, frac = frac)) - (H0.log_marginal(y) - H0.weighted_log_marginal(y, weights = frac[z_flat]))
    return np.exp(log_bf10)


import numpy as np
from numpy.testing import *
from vbayesfa.fit_lpa import fit_lpa
from vbayesfa import simulate_data

# to run all tests: pytest /Users/sampaskewitz/Documents/vbayesfa/src/vbayesfa/tests

def test_fit_lpa():
    # create simple synthetic data
    sim_data = simulate_data.latent_profile(n = 20, 
                                            m = 5,
                                            separation = 3.0,
                                            n_profiles = 2,
                                            sample_z = True,
                                            seed = 1234,
                                            mu = None,
                                            pi = None)

    # fit the model
    model = fit_lpa(sim_data['x'], 
                    prior_w1 = 3.0,
                    prior_w2 = 1.0,
                    prior_lambda_mu = 0.5,
                    prior_strength_for_xi = 10.0,
                    max_iter = 50, 
                    min_iter = 30, 
                    restarts = 4, 
                    T = 20, 
                    tolerance = 1e-5, 
                    seed = 1234)['best_model']
    model.print_summary()
    assert_equal(model.z_hat, np.array([0, 0, 1, 0, 0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 1, 1]))
    assert_almost_equal(model.sorted_m_mu[:,0], np.array([1.49505099, 1.27884819, -2.37783417, -2.19846618, -1.26139709]))
    assert_almost_equal(model.sorted_m_mu[:,1], np.array([1.01461892, 4.74435739, 1.54366584, 0.70047062, 2.47352321]))
    assert_almost_equal(model.elbo_list[-1], -116.39303540575003)


import numpy as np
from vbayesfa.mixture_model import mixture_model
from scipy.special import digamma, loggamma
from scipy.stats import beta as beta_dist
from scipy.stats import gamma as gamma_dist

class lpa_model(mixture_model):
    '''
    Bayesian latent profile analysis (LPA) model with a Dirichlet process prior to
    allow for a potentially unbounded number of latent profiles.
    
    Latent profiles have different means for each variable but share a common precision (xi).
    These have a normal-gamma prior.
    
    The concentration parameter (alpha) is learned.
    
    Attributes
    ----------
    * Data and related *
    
    x: array
        Matrix of observations (rows are variables and columns are participants/observations).
    T: integer
        Truncation level for the variational approximation. Represents the maximum number of
        latent profiles that can be discovered.
    index: array or None
        Participant IDs.
    x_names: ordered categorical series
        Names of observed variables.
    mask: array
        Indicates missing observations (= 1 if not missing, = 0 if missing).
    n: int
        Number of observations.
    n_per_var: array of int
        Number of observations for each variable (accounting for missing data).
    m: int
        Number of observed variables.
    seed: int
        Random seed.
    rng: random generator
        Generator for random numbers.
    
    * Prior distribution hyperparameters *
    
    prior_alpha_xi: float
        Prior conventional hyperparameter for xi.
    prior_beta_xi: float
        Prior conventional hyperparameter for xi.
    prior_m_mu: float
        Prior conventional hyperparameter for mu.
    prior_lambda_mu: float
        Prior conventional hyperparameter for mu.
    prior_tau1: float
        Prior natural hyperparameter for mu and xi.
    prior_tau2: float
        Prior natural hyperparameter for mu and xi.
    prior_tau3: float
        Prior natural hyperparameter for mu and xi.
    prior_tau4: float
        Prior natural hyperparameter for mu and xi.
    h_prior_tau: float
        Prior log-normalizer for each of the observed variables.
    prior_w1: float
        First prior hyperparameter on alpha.
    prior_w2: float
        Second prior hyperparameter on alpha.
    
    * Lists for storing the ELBO and its components at each iteration *
    
    E_log_lik_list: list
        List of expected log-likelihood values across updates.
    E_log_prior_list: list
        List of expected log-prior values across updates.
    entropy_list: list
        List of variational distribution entropy values across updates.
    elbo_list: list
        List of ELBO (evidence lower bound) values across updates.
    
    * Variational hyperparameters and related quantities *
    
    phi: array
        Latent profile membership probabilities.
    v_mu: array
        Variance of mu.
    E_xi_mu: array
        Expected value of xi*mu.
    E_xi_mu2: array
        Expected value of xi*mu^2.
    E_xi: array
        Expected value of xi.
    E_log_xi: array
        Expected value of log(xi).
    w1: float
        First hyperparameter of alpha.
    w2: float
        Second hyperparameter of alpha.
    E_alpha: float
        Expected value of alpha.
    E_log_alpha: float
        Expected value of log(alpha).
    gamma1: array
        First hyperparameter of V.
    gamma2: array
        Second hyperparameter of V.
    E_log_pi: array
        Expected value of log(pi).
    E_pi: array
        Expected value of pi.
    
    * Results arranged for interpretation and analysis *
    
    n_profiles: int
        Number of non-empty profiles (based on hard profile assignment).
    profile_order: array
        Order of profiles by number of members, i.e. profile size (based on hard profile assignment).
    sorted_phi: array
        Version of phi (profile membership probabilities) sorted by profile size
        and omitting empty profiles.
    sorted_m_mu: array
        Version of m_mu (posterior mean of mu) sorted by profile size and omitting empty profiles.
    sorted_v_mu: array
        Version of v_mu (posterior variance of mu) sorted by profile size and omitting empty profiles.
    z_hat: array
        Point estimates of z (profile membership) based on hard assignment, with profiles labeled in size order.
    z_hat_1hot: array
        The same information as z_hat, but in 1-hot encoding, i.e. z_hat_1hot[j,i] = 1 if person/observation i
        is in profile j and = 0 otherwise.
    n_per_profile: array
        Profile size, i.e. the number of members of each profile based on hard assignment (ordered by profile size).
    prop_per_profile: array
        Similar to n_per_profile, but gives proportions rather than numbers.
    
    Notes
    -----
    This is based on Blei and Jordan (2006).
    
    The concentration parameter is given a Gamma(3, 1) prior and learned via variational
    Bayes as described in Appendix B.
    '''
    def __init__(self, x, T = 20, prior_w1 = 3.0, prior_w2 = 1.0, prior_lambda_mu = 0.5, prior_strength_for_xi = 10.0, phi = None, seed = 1234):
        '''
        Parameters
        ----------
        x: data frame or array
            Observed data (data frame or matrix).  If a matrix then rows are variables and columns
            are observations; if a data frame then rows are observations and columns are variables.
            Missing values are replaced with 0 for computational reasons.
        T: integer, optional
            Truncation level for the variational approximation. Represents the maximum number of
            latent profiles that can be discovered.
        prior_w1: float, optional,
            First prior hyperparameter on alpha.
        prior_w2: float, optional
            Second prior hyperparameter on alpha.
        prior_lambda_mu: float, optional
            Controls the width of the prior distribution on mu: 
            higher values -> tighter prior around 0.
        prior_strength_for_xi: float, optional
            Controls the strength of the gamma prior distribution on xi.
            This prior is assumed to have a mean of 1, with alpha = prior_strength_for_xi/2
            and beta = prior_strength_for_xi/2.
        phi: array or None, optional
            If None (the default) then profile membership probabilities (phi) are randomly
            sampled from a Dirichlet(1, 1, ...) distribution. Otherwise this is an array
            of starting values for phi, with a number columns equal to the number of 
            observations and number of rows equal to or less than T.
        seed: integer, optional
            The seed for random number generation.
            
        Notes
        -----
        This is based on Blei and Jordan (2006).

        The concentration parameter (alpha) is given a Gamma(prior_w1, prior_w2) prior and learned via variational
        Bayes as described in Appendix B.
        
        Thus the prior mean of alpha is prior_w1/prior_w2.

        The means (mu) and precisions (xi) have a multi-mean normal-gamma prior.
        
        The fact that the prior alpha_xi = prior beta_xi implies that the prior distribution on xi
        has a mean of 1.
        '''
        super().__init__(x = x, T = T, prior_w1 = prior_w1, prior_w2 = prior_w2, prior_lambda_mu = prior_lambda_mu, prior_strength_for_xi = prior_strength_for_xi, phi = phi, seed = seed)
    
    def _initialize_q_pi(self, prior_w1, prior_w2, **other_prior_hpar):
        '''
        Initialize the variational distribution of latent class/profile base rates (pi),
        broken into stick-breaking lengths (V) and the Dirichlet process concentration
        parameter (alpha).
        '''
        # initialize distribution of the Dirichlet process concentration parameter (alpha)
        self.prior_w1 = prior_w1
        self.prior_w2 = prior_w2
        self.w1 = self.prior_w1 + self.T - 1
        self.w2 = self.prior_w2
        self.E_alpha = self.w1/self.w2
        self.E_log_alpha = digamma(self.w1) - np.log(self.w2)
        # initialize distribution of stick-breaking lengths (V)
        self.gamma1 = np.zeros(self.T)
        self.gamma2 = np.zeros(self.T)
        self.E_log_V = np.zeros(self.T)
        self.E_log_1minusV = np.zeros(self.T)
        self.E_log_pi = np.zeros(self.T)
        # update distribution of V
        self._update_q_V()
    
    def _update_q_V(self):
        # update distribution of stick-breaking lengths (V)
        for t in range(self.T):
            self.gamma1[t] = 1 + np.sum(self.phi[t,:])
            self.gamma2[t] = self.E_alpha + np.sum(self.phi[range(t+1,self.T),:])
            self.E_log_V[t] = digamma(self.gamma1[t]) - digamma(self.gamma1[t] + self.gamma2[t])
            self.E_log_1minusV[t] = digamma(self.gamma2[t]) - digamma(self.gamma1[t] + self.gamma2[t]) 
        for t in range(self.T):
            self.E_log_pi[t] = self.E_log_V[t] + np.sum(self.E_log_1minusV[range(t-1+1)])
        self.E_V = self.gamma1/(self.gamma1 + self.gamma2)
        product_term = np.concatenate([np.array([1]), np.cumprod(1 - self.E_V)[range(self.T-1)]])
        self.E_pi = self.E_V*product_term
        
    def _update_q_alpha(self):
        # update distribution of the Dirichlet process concentration parameter (alpha)
        self.w2 = self.prior_w2 - np.sum(self.E_log_1minusV[range(self.T - 1)]) # self.w1 stays constant, so we only update self.w2
        self.E_alpha = self.w1/self.w2
        self.E_log_alpha = digamma(self.w1) - np.log(self.w2)
    
    def _update_q_pi(self):
        '''
        Update the variational distribution of latent class/profile base rates (pi),
        broken into stick-breaking lengths (V) and the Dirichlet process concentration
        parameter (alpha).
        '''
        self._update_q_V()
        self._update_q_alpha()
    
    def _H_q_pi(self):
        '''
        Entropy of the variational distribution of latent class/profile base rates (pi),
        broken into stick-breaking lengths (V) and the Dirichlet process concentration
        parameter (alpha).
        '''
        H_q_V = beta_dist.entropy(self.gamma1, self.gamma2).sum()
        H_q_alpha = gamma_dist.entropy(a = self.w1, scale = 1/self.w2)
        return H_q_V + H_q_alpha
    
    def _E_log_prior_pi(self):
        '''
        Variational expectation of the log-prior on latent class/profile base rates (pi),
        broken into stick-breaking lengths (V) and the Dirichlet process concentration
        parameter (alpha).
        '''
        E_log_prior_V = np.sum((self.E_alpha - 1)*self.E_log_1minusV + self.E_log_alpha)
        E_log_prior_alpha = -self.w2*self.E_alpha + (self.w1 - 1)*self.E_log_alpha - loggamma(self.w1) + self.w1*np.log(self.w2)
        return E_log_prior_V + E_log_prior_alpha


import numpy as np
from vbayesfa.mixture_model import mixture_model
from scipy.special import digamma, loggamma
from scipy.stats import dirichlet as dirichlet_dist

class finite_lpa_model(mixture_model):
    '''
    Bayesian latent profile analysis (LPA) model with a finite number of latent profiles that
    have a Dirichlet distribution prior with fixed concentration parameter (alpha)
    
    Latent profiles have different means for each variable but share a common precision (xi).
    These have a normal-gamma prior.
    
    ** UPDATE BELOW AS NEEDED **
    
    Attributes
    ----------
    * Data and related *
    
    x: array
        Matrix of observations (rows are variables and columns are participants/observations).
    T: integer
        The number of latent profiles.
    index: array or None
        Participant IDs.
    x_names: ordered categorical series
        Names of observed variables.
    mask: array
        Indicates missing observations (= 1 if not missing, = 0 if missing).
    n: int
        Number of observations.
    n_per_var: array of int
        Number of observations for each variable (accounting for missing data).
    m: int
        Number of observed variables.
    seed: int
        Random seed.
    rng: random generator
        Generator for random numbers.
    
    * Prior distribution hyperparameters *
    
    prior_alpha_xi: float
        Prior conventional hyperparameter for xi.
    prior_beta_xi: float
        Prior conventional hyperparameter for xi.
    prior_m_mu: float
        Prior conventional hyperparameter for mu.
    prior_lambda_mu: float
        Prior conventional hyperparameter for mu.
    prior_tau1: float
        Prior natural hyperparameter for mu and xi.
    prior_tau2: float
        Prior natural hyperparameter for mu and xi.
    prior_tau3: float
        Prior natural hyperparameter for mu and xi.
    prior_tau4: float
        Prior natural hyperparameter for mu and xi.
    h_prior_tau: float
        Prior log-normalizer for each of the observed variables.
    prior_w1: float
        First prior hyperparameter on alpha.
    prior_w2: float
        Second prior hyperparameter on alpha.
    
    * Lists for storing the ELBO and its components at each iteration *
    
    E_log_lik_list: list
        List of expected log-likelihood values across updates.
    E_log_prior_list: list
        List of expected log-prior values across updates.
    entropy_list: list
        List of variational distribution entropy values across updates.
    elbo_list: list
        List of ELBO (evidence lower bound) values across updates.
    
    * Variational hyperparameters and related quantities *
    
    phi: array
        Latent profile membership probabilities.
    v_mu: array
        Variance of mu.
    E_xi_mu: array
        Expected value of xi*mu.
    E_xi_mu2: array
        Expected value of xi*mu^2.
    E_xi: array
        Expected value of xi.
    E_log_xi: array
        Expected value of log(xi).
    
    ** ADD HYPERPARAMETERS FOR PI **
    
    E_log_pi: array
        Expected value of log(pi).
    E_pi: array
        Expected value of pi.
    
    * Results arranged for interpretation and analysis *
    
    n_profiles: int
        Number of non-empty profiles (based on hard profile assignment).
    profile_order: array
        Order of profiles by number of members, i.e. profile size (based on hard profile assignment).
    sorted_phi: array
        Version of phi (profile membership probabilities) sorted by profile size
        and omitting empty profiles.
    sorted_m_mu: array
        Version of m_mu (posterior mean of mu) sorted by profile size and omitting empty profiles.
    sorted_v_mu: array
        Version of v_mu (posterior variance of mu) sorted by profile size and omitting empty profiles.
    z_hat: array
        Point estimates of z (profile membership) based on hard assignment, with profiles labeled in size order.
    z_hat_1hot: array
        The same information as z_hat, but in 1-hot encoding, i.e. z_hat_1hot[j,i] = 1 if person/observation i
        is in profile j and = 0 otherwise.
    n_per_profile: array
        Profile size, i.e. the number of members of each profile based on hard assignment (ordered by profile size).
    prop_per_profile: array
        Similar to n_per_profile, but gives proportions rather than numbers.
    '''
    def __init__(self, x, T = 4, prior_alpha_pi = 0.5, prior_lambda_mu = 0.5, prior_strength_for_xi = 10.0, phi = None, seed = 1234):
        '''
        Parameters
        ----------
        x: data frame or array
            Observed data (data frame or matrix).  If a matrix then rows are variables and columns
            are observations; if a data frame then rows are observations and columns are variables.
            Missing values are replaced with 0 for computational reasons.
        T: integer, optional
            The number of latent profiles.
        prior_alpha_pi: float, optional
            Prior strength of the Dirichlet prior distribution on pi.
        prior_lambda_mu: float, optional
            Controls the width of the prior distribution on mu: 
            higher values -> tighter prior around 0.
        prior_strength_for_xi: float, optional
            Controls the strength of the gamma prior distribution on xi.
            This prior is assumed to have a mean of 1, with alpha = prior_strength_for_xi/2
            and beta = prior_strength_for_xi/2.
        phi: array or None, optional
            If None (the default) then profile membership probabilities (phi) are randomly
            sampled from a Dirichlet(1, 1, ...) distribution. Otherwise this is an array
            of starting values for phi, with a number columns equal to the number of 
            observations and number of rows equal to or less than T.
        seed: integer, optional
            The seed for random number generation.
            
        Notes
        -----
        The means (mu) and precisions (xi) have a multi-mean normal-gamma prior.
        
        The fact that the prior alpha_xi = prior beta_xi implies that the prior distribution on xi
        has a mean of 1.
        '''
        super().__init__(x = x, T = T, prior_alpha_pi = prior_alpha_pi, prior_lambda_mu = prior_lambda_mu, prior_strength_for_xi = prior_strength_for_xi, phi = phi, seed = seed)
    
    def _initialize_q_pi(self, prior_alpha_pi, **other_prior_hpar):
        '''
        Initialize the variational distribution of latent class/profile base rates (pi).
        '''
        self.prior_alpha_pi = prior_alpha_pi
        self.alpha_pi = self.prior_alpha_pi*np.ones(self.T)
        self._update_q_pi()
    
    def _update_q_pi(self):
        '''
        Update the variational distribution of latent class/profile base rates (pi).
        '''
        self.alpha_pi = self.prior_alpha_pi + self.phi.sum(axis = 1)
        self.E_pi = self.alpha_pi/self.alpha_pi.sum()
        self.E_log_pi = digamma(self.alpha_pi) - digamma(self.alpha_pi.sum())
    
    def _H_q_pi(self):
        '''
        Entropy of the variational distribution of latent class/profile base rates (pi).
        '''
        return dirichlet_dist.entropy(self.alpha_pi)
    
    def _E_log_prior_pi(self):
        '''
        Variational expectation of the log-prior on latent class/profile base rates (pi).
        '''
        return np.sum((self.alpha_pi - 1)*self.E_log_pi) - np.sum(loggamma(self.alpha_pi)) + loggamma(self.alpha_pi.sum())  


from .fit_mixture_model import fit_mixture_model
from .finite_lpa import finite_lpa_model

def fit_finite_lpa(x, n_workers = 4, kmeans = False, restarts = 20, T = 20, prior_alpha_pi = 1.0, prior_lambda_mu = 0.5, prior_strength_for_xi = 10.0, seed = 1234, tolerance = 1e-05, max_iter = 1000, min_iter = 10):
    """
    Fit a finite latent profile analysis (DPM-LPA) model using mean field variational Bayes.
    
    Parameters
    ----------
    x: data frame or array
        Observed data (data frame or matrix).  If a matrix then rows are variables and columns
        are observations; if a data frame then rows are observations and columns are variables.
    n_workers: int, optional
        Number of workers in the pool for parallel processing; should be less than
        or equal to the number of available CPUs.
    kmeans: logical, optional
        Should the k-means++ algorithm be used for initial starting points?
        If so, then k-means clusterings of different sizes are produced and used.
    restarts: int, optional
        Number of random restarts per value of T (number of latent profiles).
    T: integers, optional
        Number of latent profiles (some may end up empty).
    prior_alpha_pi: float, optional
        Prior strength of the Dirichlet prior distribution on pi.
    prior_lambda_mu: float, optional
        Controls the width of the prior distribution on mu: 
        higher values -> tighter prior around 0.
    prior_strength_for_xi: float, optional
        Controls the strength of the gamma prior distribution on xi.
        This prior is assumed to have a mean of 1, with alpha = prior_strength_for_xi/2
        and beta = prior_strength_for_xi/2.
    seed: int, optional
        Random seed (determines starting conditions of each optimization).
    tolerance: float, optional
        Relative change in the ELBO at which the optimization should stop.
    max_iter: integer, optional
        Maximum number of iterations to run the optimization.
    min_iter: integer, optional
        Minimum number of iterations to run the optimization.
    
    Notes
    -----
    This is a wrapper for the fit_mixture_model function.
    
    One fits a statistical model using variational Bayes by maximizing a quantity called
    the ELBO (evidence lower bound). The best variational Bayes approximation of the true
    model posterior distribution is the one with the highest ELBO. The algorithm for maximizing
    the ELBO unfortunately only finds local maxima, so to have a decent chance of finding the
    global maximum (truly best approximation to the real model posterior distribution) we need
    different starting points of the variational model hyperparameters. This function provides two
    ways to choose starting points. First, the k-means++ algorithm (a simple clustering algorithm)
    can be used to produce initial clustering of participants. Second, participants can be randomly
    assigned (in a 'soft' way) to different mixture components/latent classes/latent profiles. These
    two methods are not exclusive: both types of starting point will be used if 'kmeans' is set to
    True and 'restarts' > 0. The function creates a separate mixture model object with each starting
    point and fits them using parallel processing for speed.    
    
    Output
    ------
    A dictionary with the following:
    
    final_elbo: list
        Final ELBO (evidence lower bound) values of each model.
    model_list: list
        Fitted model objects.
    best_model: mixture_model object
        Best fitted model (i.e. the one with the highest ELBO).
    fit_time: float
        Total time used to fit.
    """
    return fit_mixture_model(x = x, model_class = finite_lpa_model, kmeans = kmeans, n_workers = n_workers, restarts = restarts, T = T, prior_alpha_pi = prior_alpha_pi, prior_lambda_mu = prior_lambda_mu, prior_strength_for_xi = prior_strength_for_xi, seed = seed, tolerance = tolerance, max_iter = max_iter, min_iter = min_iter)